# 1st party
import typing
from abc import ABC, abstractmethod

# 3rd party
import torch.nn as nn

# local
from .core import ID, UNDEFINED
from .nodes import End, In, Info, Layer, Process, ProcessSet, is_defined, to_incoming


class Tako(nn.Module):
    """Base class for Takos which wrap processses to make more flexible networks"""

    @abstractmethod
    def forward_iter(self, in_: Process = None) -> typing.Iterator:
        """Method that loops over each process one by one

        Args:
            in_ (Process, optional): _description_. Defaults to None.

        Returns:
            typing.Iterator: An iterator over the processes
        Yields:
            Process: Each process in the Tako

        """
        pass

    def sub(
        self,
        y: typing.Union[str, typing.List[str]],
        x: typing.Union[str, typing.List[str]],
    ):
        """
        Extract a sub network form the Tako

        TODO: Simplify the code
        """

        x_is_list = isinstance(x, list)
        y_is_list = isinstance(y, list)

        if y_is_list:
            out = {y_i: UNDEFINED for y_i in y}
        else:
            out = UNDEFINED

        if x_is_list:
            found = [False] * len(x)
        else:
            found = False

        in_ = In()

        for layer in self.forward_iter(in_):
            if x_is_list and layer.name in x:
                idx = x.index(layer.name)
                in_[idx].rewire(layer)
                found[idx] = True
            elif not x_is_list and layer.name == x:
                in_.rewire(layer)
                found = True

            if y_is_list and layer.name in y:
                out[layer.name] = layer

        if (x_is_list and False in found) or (not x_is_list and found is False):
            raise RuntimeError()

        if (y_is_list and UNDEFINED in out.values()) or (
            not y_is_list and out is UNDEFINED
        ):
            raise RuntimeError()
        elif y_is_list:
            return list(out.values())
        return out

    def probe(
        self,
        y: typing.Union[str, typing.List[str]],
        in_: Process = None,
        by: typing.Dict[str, typing.Any] = None,
    ):
        """Probe the output of a module based on the input

        Args:
            y (typing.Union[str, typing.List[str]]): The Processes to probe
            in_ (Process, optional): The incoming process. Defaults to None.
            by (typing.Dict[str, typing.Any], optional): The inputs to the processes. Defaults to None.

        Returns:
            Any: the result of the probe
        """
        by = by or {}
        if isinstance(y, list):
            out = {y_i: UNDEFINED for y_i in y}
            is_list = True
        else:
            is_list = False
            out = UNDEFINED

        for layer in self.forward_iter(in_):
            for id, x in by.items():
                if layer.name == id:
                    layer.y = x

            if is_list and layer.name in out:
                out[layer.name] = layer.y
            elif not is_list:
                if layer.name == y:
                    return layer.y

        if is_list:
            return list(out.values())
        return out

    def forward(self, x) -> typing.Any:
        """Method to

        Args:
            x (Input to the first process):

        Returns:
            Any: The output of the Tako
        """
        y = x
        for layer in self.forward_iter(In(x)):
            y = layer.y
        return y

    def from_(
        self, process: "Process", name=None, info=None, dive: bool = True
    ) -> typing.Iterator["Process"]:
        """_summary_

        Yields:
            _type_: _description_
        """

        # TODO: Brainsorm more about this
        if dive is None:
            yield Nested(self, x=to_incoming(process), name=name, info=info)
        else:
            nested = Nested(self, x=to_incoming(process), name=name, info=info)
            for layer in nested.y_iter():
                yield layer


class Sequence(Tako):
    """Tako consisting of a sequence of nodes"""

    def __init__(self, modules: typing.Iterable[nn.Module]):
        super().__init__()
        self._sequence = nn.ModuleList(modules)

    def forward_iter(self, in_: Process):

        cur = in_ or In()
        for module in self._sequence:
            cur = cur.to(module)
            yield cur


class Filter(ABC):
    """Base class for filtering the layers in a Tako"""

    @abstractmethod
    def check(self, layer: Layer) -> bool:
        pass

    def extract(self, tako: Tako) -> ProcessSet:
        return ProcessSet([layer for layer in self.filter(tako) if self.check(layer)])

    def apply(self, tako: Tako, process: Process):
        for layer in self.filter(tako):
            if self.check(layer):
                process.apply(layer)

    def filter(self, tako: "Tako") -> typing.Iterator:
        for layer in tako.forward_iter():
            if self.check(layer):
                yield layer


class TagFilter(Filter):
    """Filter the Tako by a tag"""

    def __init__(self, filter_tags: typing.List[str]):

        self._filter_tags = set(filter_tags)

    def check(self, layer: Layer) -> bool:
        return len(self._filter_tags.intersection(layer.info.tags)) > 0


def layer_dive(layer: Layer) -> typing.Iterator[Process]:
    """Loop over all sub layers in a layer including

    Args:
        layer (Layer): _description_

    Returns:
        typing.Iterator[Process]: Iterator

    Yields:
        the layers: _description_
    """
    if isinstance(layer, Nested):
        for layer_i in layer.tako.forward_iter():
            for layer_j in layer_dive(layer_i):
                yield layer_j

    else:
        yield layer_i


def dive(tako: Tako, in_: Process) -> typing.Iterator[Process]:
    """Loop over the processes in the Tako

    Args:
        tako (Tako): The Tako to loop over
        in_ (Node): the incoming process

    Returns:
        typing.Iterator[Process]: The iterator

    Yields:
        Process: Process in the layer
    """
    for layer in tako.forward_iter(in_):
        yield layer_dive(layer)


def processes(tako: Tako, in_) -> typing.List[Process]:
    """Get the processes for a Tako

    Args:
        tako (Tako): The Tako to get the processes for
        in_ (_type_): The incoming nodes

    Returns:
        typing.List[Process]:
    """
    outs = []
    curs = []
    for layer in tako.forward_iter(in_):
        if isinstance(layer, End):
            outs.append(curs)
            curs = []
            curs.append(layer)
        else:
            curs.append(layer)
    if len(curs) != 0:
        outs.append(curs)
    return outs


class Nested(Process):
    """Tako nested inside another Tako"""

    # TODO: Consider the design of this
    # to allow for nesting processes.. I want to
    # be able to use forward_iter

    # 1) if the input is UNDEFINED it will output undefined and
    # will not execute the underlying processes
    # 2) if the input is not UNDEFINED it will iterate over
    # all the underlying processes.. So for a nested process I
    #

    #

    def __init__(self, tako: "Tako", x=UNDEFINED, name: str = None, info: Info = None):
        """initializer

        Args:
            nn_module (typing.Union[typing.List[nn.Module], nn.Module]): Module for the layer
            x (optional): Input to the node. Defaults to UNDEFINED.
            name (str, optional): Name of the node. Defaults to None.
            info (Info, optional): Info for the node. Defaults to None.
        """
        super().__init__(x, name=name, info=info)
        self.tako = tako
        self._y = UNDEFINED
        self._outs = None
        self._in = None

    @property
    def y(self) -> typing.Any:
        """
        Returns:
            typing.Any: The output of the process
        """

        if self._y == UNDEFINED and isinstance(self._x, Process):
            return UNDEFINED

        elif self._y == UNDEFINED and self._x != UNDEFINED:
            self._in = In(self._x)

            self._outs = processes(self.tako, self._in)

        self._y = self._outs[-1][-1].y
        return self._y

    @y.setter
    def y(self, y):
        """
        Args:
            y (): The output of the node
        """
        self._y = y

    def _probe_out(self, by):
        if is_defined(self._x):
            x = self._x
        else:
            x = self._x.probe(by)
        return self.tako(x)

    def y_iter(self) -> typing.Iterator[Process]:
        """Iterate over the Tako

        Returns:
            typing.Iterator[Process]:

        Yields:
            typing.Any
        """

        outs = []
        for process in self.tako.forward_iter(self._x):
            yield process
            outs.append(process)
        self._outs = outs
        self._y = self._outs[-1]


class Network(nn.Module):
    """A wrapping an 'in-node' and an 'out-node'"""

    def __init__(
        self,
        out: typing.Union[Process, ProcessSet],
        in_: typing.Union[ID, typing.List[str]],
        by,
    ):
        """_summary_

        Args:
            out (typing.Union[Process, ProcessSet]): _description_
            in_ (typing.Union[ID, typing.List[str]]): _description_
            by (_type_): _description_
        """
        self._out = out
        self._in = in_
        self._by = by

    def forward(self, x):

        # need to mark which ones
        # i want to store in by
        # by = by.update(**self._in, x)
        # by.outgoing_count(t)

        by = {**self._by, **zip(self._in, x)}
        return self._out.probe(by)



# 1st party
import typing
from abc import abstractproperty


class IDable:
    @abstractproperty
    def id(self) -> str:
        pass


class StateKeyError(KeyError):
    """Used to make errors that state returns more explicit"""

    pass


class State(object):

    # add in x and y by default
    # so that XOptim, ThetaOptim do not have
    # to be coupled to the lm

    def __init__(self):
        super().__init__()
        self._data = {}
        self._subs = {}

    def id(self, obj):
        if isinstance(obj, str):
            return obj

        try:
            return obj.id
        except AttributeError:
            return id(obj)

    def store(self, obj: IDable, key: typing.Hashable, value):

        obj_data, _ = self._get_obj(obj)
        obj_data[key] = value
        return value

    def get(self, obj: IDable, key: typing.Hashable, default=None):
        try:
            if isinstance(key, typing.List):
                return [self._data[self.id(obj)][key_i] for key_i in key]
            else:
                return self._data[self.id(obj)][key]
        except KeyError:
            return default

    def __getitem__(self, index: typing.Tuple[IDable, typing.Hashable]):
        obj, key = index
        obj_id = self.id(obj)
        try:
            if isinstance(key, typing.List):
                return [self._data[obj_id][key_i] for key_i in key]
            else:
                return self._data[obj_id][key]
        except KeyError:
            raise StateKeyError(
                f"There is no recorded state for {obj} of type {type(obj)} and key {key}"
            )

    def __setitem__(self, index: typing.Tuple[IDable, typing.Hashable], value):

        obj, key = index
        return self.store(obj, key, value)

    def _get_obj(self, obj, to_add: bool = True):
        id = self.id(obj)
        if to_add and id not in self._data:
            self._data[id] = {}
            self._subs[id] = {}
        return self._data[id], self._subs[id]

    def add_sub(self, obj: IDable, key: str, ignore_exists: bool = True) -> "State":
        """Add a 'sub state' specified by key

        Args:
            key (str): The key for the substate
            ignore_exists (bool, optional): Whether to ignore if substate already exists. Defaults to True.

        Raises:
            KeyError: If ignore exists if false and key already exists

        Returns:
            State: The substate created
        """
        _, sub_data = self._get_obj(obj)

        if key in sub_data:
            raise StateKeyError(
                f"Subs State {key} is already in State and ignore exists is False."
            )
        result = sub_data[key] = State()
        return result

    def my_sub(self, obj: IDable, key: str, to_add: bool = True) -> "MyState":
        mine = self.mine(obj, to_add)
        return mine.my_sub(key, to_add)

    def sub(self, obj: IDable, key: str, to_add: bool = True) -> "State":
        """Retrieve a sub state

        Args:
            key (str): The name of the sub state
            to_add (bool, optional): Whether to add the state if it does not exist. Defaults to True.

        Raises:
            KeyError: If the sub state does not exist and to_add is false

        Returns:
            State: The substate
        """
        _, sub_data = self._get_obj(obj, to_add)
        if to_add and key not in sub_data:
            state = sub_data[key] = State()
            return state
        return sub_data[key]

    def clear(self, obj: IDable):
        id = self.id(obj)
        if id in self._data:
            self._data[id].clear()
        if id in self._subs:
            self._subs[id].clear()

    def sub_iter(self, obj) -> typing.Iterator[typing.Tuple[str, "State"]]:
        """Iterator over all sub states

        Yields:
            typing.Iterator[typing.Tuple[str, 'State']]: The name and state of all substates
        """
        id = self.id(obj)
        for key, value in self._subs[id].items():
            yield key, value

    def mine(self, obj, to_add: bool = True) -> "MyState":
        return MyState(obj, *self._get_obj(obj, to_add=to_add))

    def __contains__(self, key):
        obj, key = key
        id = self.id(obj)
        return id in self._data and key in self._data[id]


class MyState(object):
    """Convenience class to make it easy to access state values for an object
    x = object()
    my_state = State().mine(x)
    my_state.t = 1
    """

    def __init__(self, obj, data: typing.Dict, subs: typing.Dict):
        """initializer

        Args:
            data (typing.Dict): The data for the object
            subs (typing.Dict): The sub states for the object
        """
        object.__setattr__(self, "_obj", obj)
        object.__setattr__(self, "_data", data)
        object.__setattr__(self, "_subs", subs)

    @property
    def subs(self) -> typing.Dict[str, State]:
        return self._subs

    def add_sub(self, key: str, state: State = None) -> "State":
        """Add a substate to the state

        Args:
            key (str): Name of the sub state
            state (State): The substate to add

        Raises:
            KeyError: The key already exists in the sub states
        """
        if key in self._subs:
            raise KeyError(f"State by name of {key} already exists in subs")
        state = state or State()
        self._subs[key] = state
        return state

    def my_sub(self, key: str, to_add: bool = True) -> "MyState":
        if to_add and key not in self._subs:
            self._subs[key] = State()
        return self._subs[key].mine(self._obj)

    def get(self, key: str, default=None):
        return self._data.get(key, default)

    def __getitem__(self, key):
        return self._data[key]

    def __setitem__(self, key, value):
        self._data[key] = value

    def __getattr__(self, key):
        return self._data[key]

    def __setattr__(self, key: str, value: typing.Any) -> typing.Any:
        self._data[key] = value
        return value

    def __contains__(self, key: str) -> bool:
        return key in self._data



# 1st party
import typing
from abc import ABC, abstractmethod

# local
from ..kaku import IO, LearningMachine
from .core import Population, expand_t, reduce_assessment_dim1


class PopulationAssessor(ABC):
    """Modules to asseess the population"""

    @abstractmethod
    def assess(self, population: Population, t: IO) -> Population:
        pass

    def __call__(self, population: Population, t: IO) -> Population:
        return self.assess(population, t)


class XPopulationAssessor(PopulationAssessor):
    """Assess the inputs to the population"""

    def __init__(
        self,
        learner: LearningMachine,
        names: typing.List[str],
        loss_name: str,
        reduction: str,
        k: int,
    ):

        self.learner = learner
        self.names = names
        self.reduction = reduction
        self.loss_name = loss_name
        self.k = k

    def assess(self, population: Population, t: IO) -> Population:

        t = expand_t(t, len(population))

        assessment = self.learner.assess(
            IO(*population.flattened(self.names)), t, reduction_override="none"
        )[self.loss_name]
        assessment = reduce_assessment_dim1(assessment, self.k, True)
        population.report(assessment)

        return population



# 1st party
import typing
from abc import ABC, abstractmethod, abstractproperty

# 3rd party
import torch.nn as nn

# local
from .core import UNDEFINED, Info


def is_defined(x):
    """Whether an input is defined"""
    return not isinstance(x, Process) and x != UNDEFINED


def get_x(node):
    """Get the value of x"""
    if isinstance(node.x, Process):
        return UNDEFINED
    return node.x


def to_incoming(node):
    """Return y if node is undefined otherwise return the node"""

    if node.y is UNDEFINED:
        return node
    return node.y


class Process(ABC):
    """Base class for all network nodes"""

    def __init__(self, x=UNDEFINED, name: str = None, info: Info = None):
        """initializer

        Args:
            x (optional): The input to the node. Defaults to UNDEFINED.
            name (str, optional): The name of the node. Defaults to None.
            info (Info, optional): Infor for the node. Defaults to None.
        """
        self._name = name or str(id(self))
        self._info = info or Info()
        self._outgoing = []
        self._x = UNDEFINED
        self.x = x

    @property
    def name(self) -> str:
        """

        Returns:
            str: Name of the ndoe
        """

        return self._name

    @name.setter
    def name(self, name: str):
        self._name = name

    @abstractproperty
    def y(self):
        """
        Returns:
            Process output if defined else Undefined
        """
        raise NotImplementedError

    @y.setter
    def y(self, y):
        """
        Args:
            y (): The output for the node
        """
        raise NotImplementedError

    @property
    def x(self):
        """

        Returns:
             The input into the network
        """
        if isinstance(self._x, Process):
            return UNDEFINED
        return self._x

    @x.setter
    def x(self, x):
        """
        Args:
            x (): Set the input to the network
        """
        if isinstance(self._x, Process):
            self._x._remove_outgoing(self)
        if isinstance(x, Process):
            x._add_outgoing(self)
        self._x = x

    def lookup(self, by: dict):
        """
        Args:
            by (): _description_

        Returns:
            _type_: _description_
        """
        return by.get(self._name, UNDEFINED)

    def set_by(self, by, y):
        """Set the value of y

        Args:
            by (dict):
            y (): The output to the network
        """
        by[self.name] = y

    def _add_outgoing(self, process: "Process"):
        process._outgoing.append(self)

    def _remove_outgoing(self, process: "Process"):
        process._outgoing.remove(self)

    # def nest(self, tako: 'Tako', name=None, info=None, dive: bool=True) -> typing.Iterator['Process']:

    #     # TODO: Brainsorm more about this
    #     if dive is None:
    #         yield Nested(tako, x=to_incoming(self), name=name, info=info)
    #     else:
    #         nested = Nested(tako, x=to_incoming(self), name=name, info=info)
    #         for layer in nested.y_iter():
    #             yield layer

    def to(
        self,
        nn_module: typing.Union[typing.List[nn.Module], nn.Module],
        name: str = None,
        info: Info = None,
    ) -> "Layer":
        """Connect the layer to another layer

        Args:
            nn_module (typing.Union[typing.List[nn.Module], nn.Module]): Module
            name (str, optional): Name of the node. Defaults to None.
            info (Info, optional): Info for the layer. Defaults to None.

        Returns:
            Layer:
        """
        return Layer(nn_module, x=to_incoming(self), name=name, info=info)

    def join(self, *others, info: Info = None) -> "Joint":
        """Join two layers

        Args:
            *others: nodes to join with
            info (Info, optional): _description_. Defaults to None.

        Returns:
            Joint
        """

        ys = []
        for node in [self, *others]:
            ys.append(to_incoming(node))

        return Joint(x=ys, info=info)

    def get(self, idx: typing.Union[slice, int], info: Info = None):
        """Get an index from the output

        Args:
            idx (typing.Union[slice, int]):
            info (Info, optional): _description_. Defaults to None.

        Returns:
            Index
        """
        return Index(idx, x=to_incoming(self), info=info)

    def __getitem__(self, idx: int):
        return self.get(idx)

    @abstractmethod
    def _probe_out(self, by: dict):
        raise NotImplementedError

    def probe(self, by):
        """Probe the layer

        Args:
            by (dict)

        Returns:
            result: value
        """
        value = self._info.lookup(by)
        if value is not None:
            return value

        result = self._probe_out(by)

        if len(self._outgoing) > 1:
            self.set_by(by, result)
        return result


class Joint(Process):
    def __init__(self, x=UNDEFINED, name: str = None, info: Info = None):
        """initializer

        Args:
            x (_type_, optional): Input to the node. Defaults to UNDEFINED.
            name (str, optional): Name of the node. Defaults to None.
            info (Info, optional): Info for the node. Defaults to None.
        """
        super().__init__(x, name=name, info=info)
        self._y = UNDEFINED

    @property
    def y(self):

        undefined = False
        for x_i in self._x:
            if isinstance(x_i, Process):
                undefined = True

        if undefined:
            self._y = UNDEFINED
        else:
            self._y = self._x

        return self._y

    @y.setter
    def y(self, y):
        self._y = y

    @property
    def x(self):
        """
        Returns:
            The input to the node
        """

        if self._x is UNDEFINED:
            return UNDEFINED
        x = []
        for x_i in self._x:
            if isinstance(x_i, Process):
                x.append(UNDEFINED)
            else:
                x.append(x_i)

        return x

    @x.setter
    def x(self, x):

        if self._x is not UNDEFINED:
            for x_i in self._x:
                if isinstance(x_i, Process):
                    x_i._remove_outgoing(self)

        xs = []
        for x_i in x:
            if isinstance(x_i, Process):
                x_i._add_outgoing(self)
            xs.append(x_i)
        self._x = xs

    def _probe_out(self, by):
        y = []
        for x_i in self._x:
            if is_defined(x_i):
                y.append(x_i)
            else:
                y.append(x_i.probe(by))
        return y


class Index(Process):
    """

    Args:
        Process (_type_): _description_
    """

    def __init__(
        self,
        idx: typing.Union[int, slice],
        x=UNDEFINED,
        name: str = None,
        info: Info = None,
    ):
        super().__init__(x, name, info)
        self._idx = idx

    @property
    def y(self):
        """

        Returns:
            The output of the node
        """
        if isinstance(self._x, Process):
            return UNDEFINED
        else:
            return self._x[self._idx]

    def _probe_out(self, by):

        if is_defined(self._x):
            x = self._x
        else:
            x = self._x.probe(by)

        return x[self._idx]


class End(Process):
    def __init__(self, x=UNDEFINED, name: str = None, info: Info = None):
        """initializer

        Args:
            nn_module (typing.Union[typing.List[nn.Module], nn.Module]): Module for the layer
            x (optional): Input to the node. Defaults to UNDEFINED.
            name (str, optional): Name of the node. Defaults to None.
            info (Info, optional): Info for the node. Defaults to None.
        """
        super().__init__(x, name=name, info=info)
        self._y = UNDEFINED

    def y_(self, store: bool = True):
        pass

    @property
    def y(self):
        if is_defined(self._y):
            return self._y
        return self.x

    @y.setter
    def y(self, y):
        """
        Args:
            y (): The output of the node
        """
        self._y = y

    def _probe_out(self, by):
        if is_defined(self._x):
            x = self._x
        else:
            x = self._x.probe(by)
        return x


class Layer(Process):
    def __init__(
        self,
        nn_module: typing.Union[typing.List[nn.Module], nn.Module],
        x=UNDEFINED,
        name: str = None,
        info: Info = None,
    ):
        """initializer

        Args:
            nn_module (typing.Union[typing.List[nn.Module], nn.Module]): Module for the layer
            x (optional): Input to the node. Defaults to UNDEFINED.
            name (str, optional): Name of the node. Defaults to None.
            info (Info, optional): Info for the node. Defaults to None.
        """
        super().__init__(x, name=name, info=info)
        if isinstance(nn_module, typing.List):
            nn_module = nn.Sequential(*nn_module)
        self.op = nn_module
        self._y = UNDEFINED

    def y_(self, store: bool = True):
        pass

    @property
    def y(self):

        # return self._y

        if self._y == UNDEFINED and isinstance(self._x, Process):
            return UNDEFINED

        elif self._y == UNDEFINED and self._x != UNDEFINED:
            self._y = self.op(self._x)

        return self._y

    @y.setter
    def y(self, y):
        """
        Args:
            y (): The output of the node
        """
        self._y = y

    def _probe_out(self, by):
        if is_defined(self._x):
            x = self._x
        else:
            x = self._x.probe(by)
        return self.op(x)


class In(Process):
    @property
    def x(self):
        return self._x

    @x.setter
    def x(self, x):
        self._x = x

    @property
    def y(self):
        return self._x

    @y.setter
    def y(self, y):
        self._x = y

    def _probe_out(self, by):
        return self._x


class ProcessSet(object):
    """Set of nodes. Can use to probe"""

    def __init__(self, nodes: typing.List[Process]):
        self._nodes = {node.name: node for node in nodes}

    def apply(self, process: Process):
        """Apply a process on each node in the set

        Args:
            process (Process)
        """
        for node in self._nodes.values():
            if isinstance(process, Apply):
                process.apply(node)
            else:
                process(node)

    def __getitem__(self, key: str) -> Process:
        """
        Args:
            key (str): name of the nodeset

        Returns:
            Process
        """
        if key not in self._nodes:
            raise KeyError("There is no node named key.")
        return self._nodes[key]

    def probe(self, by):
        """
        Args:
            by (dict): Outputs for nodes {'node': {output}}

        Returns:
            typing.Union[typing.List[torch.Tensor], torch.Tensor]
        """
        result = []
        for node in self._nodes:
            result.append(node.probe(by))
        return result

    def __iter__(self) -> typing.Iterator[Process]:

        for process in self._nodes.values():
            yield process

    def __len__(self) -> int:
        return


class ProcessVisitor(ABC):
    """Base class for a module that visits a process"""

    @abstractmethod
    def visit(self, process: "Process"):
        pass


class Apply(ABC):
    """Apply a function to a process"""

    @abstractmethod
    def apply(self, node: "Process"):
        pass


class LambdaApply(Apply):
    """Pass in a Lambda function"""

    def __init__(self, f, *args, **kwargs):
        """initializer

        Args:
            f (_type_): The function to apply to a process
        """
        self._f = f
        self._args = args
        self._kwargs = kwargs

    def apply(self, node: Process):
        self._f(node, *self._args, **self._kwargs)



# import typing

# # 1st Party
# from abc import abstractmethod, abstractproperty
# from dataclasses import dataclass

# import numpy as np

# # 3rd Party
# import torch
# import torch.nn as nn
# import torch.nn.functional as nn_func
# from torch import nn

# # Local
# from ... import utils

# # TODO: Move as this is the wrong place
# class KLDivLoss(nn.Module):
#     """An extended version of KL divergence
#     """

#     def __init__(self, reduction: str='batchmean', class_target: bool=True, logits: bool=True):
#         super().__init__()
#         self._kl_div = nn.KLDivLoss(reduction=reduction)
#         self._inv = SigmoidInvertable()
#         self.class_target = class_target
#         self.logits = logits

#     def forward(self, x: torch.Tensor, t: torch.Tensor) -> torch.Tensor:
#         """Calculate the KL Divergence Loss allowing class inputs

#         Args:
#             x (torch.Tensor): the input tensor
#             t (torch.Tensor): the target tensor

#         Returns:
#             Tensor: the KL divergence between the input and the output
#         """
#         if self.class_target:
#             t = torch.nn.functional.one_hot(t)

#         if self.logits:
#             x = nn_func.log_softmax(x, dim=1)
#         else:
#             x = self._inv.reverse(x)

#         return self._kl_div(x, t)


# from .assess import AssessmentDict, Reduction, Assessment


# class L2Reg(nn.Module):

#     def __init__(self, lam: float=1e-1, reduction: str='mean'):
#         super().__init__()
#         self._lam = lam
#         self._reduction = reduction
#         if reduction not in ('none', 'batchmean', 'mean'):
#             raise RuntimeError(f'Reduction must be "none", "batchmean" or "mean" {reduction} ')

#     def forward(self, x: torch.Tensor):

#         if self._reduction == 'none':
#             return (x ** 2) * self._lam

#         elif self._reduction == 'batchmean':
#             return ((x ** 2).sum() * self._lam) / len(x)

#         elif self._reduction == 'mean':
#             return (x ** 2).mean() * self._lam


# class L1Reg(nn.Module):

#     def __init__(self, lam: float=1e-1, reduction: str='mean'):
#         super().__init__()
#         self._lam = lam
#         self._reduction = reduction
#         if reduction not in ('none', 'batchmean', 'mean'):
#             raise RuntimeError(f'Reduction must be "none", "batchmean" or "mean" {reduction} ')

#     def forward(self, x: torch.Tensor):

#         # TODO: Move to Losses
#         if self._reduction == 'none':
#             return torch.abs(x) * self._lam

#         elif self._reduction == 'batchmean':
#             return (torch.abs(x) * self._lam) / len(x)

#         elif self._reduction == 'mean':
#             return torch.abs(x).mean() * self._lam


# class DXModLoss(Loss):
#     """Wraps a module and a loss together so that more advanced backpropagation
#     can be implemented
#     """

#     @abstractproperty
#     def module(self) -> nn.Module:
#         raise NotImplementedError

#     @abstractmethod
#     def forward(self, x: torch.Tensor, y: torch.Tensor, dx: torch.Tensor, reduction_override=None):
#         raise NotImplementedError

#     def assess(self, x: torch.Tensor, y: torch.Tensor, dx: torch.Tensor, reduction_override: str=None):
#         return (
#           Assessment(self.forward(x, y, dx, reduction_override), 
#           self._maximize)

#     def assess_dict(
#         self, x: torch.Tensor, y: torch.Tensor, dx: torch.Tensor, 
#         reduction_override: str=None, name: str='loss'):
#         return AssessmentDict(
#             **{name: Assessment(self.forward(x, y, dx, reduction_override), self._maximize)}
#         )


# class WeightedLoss(nn.Module):

#     def __init__(self, loss: nn.Module, weight: float=1.0, reduction='mean'):
#         super().__init__()
#         self.loss = loss
#         self.weight = weight
#         self.loss.reduction = reduction

#     def forward(self, x: torch.Tensor, t: torch.Tensor):
#         return self.weight * self.loss(x, t)

#     @property
#     def reduction(self):
#         return self.loss.reduction

#     @reduction.setter
#     def reduction(self, reduction: str):
#         self.loss.reduction = reduction


# class DXLossGrad(torch.autograd.Function):
#     """Use to implement backpropagation that is closer to the standard
#     form that sends information in a direct path backward through the layers
#     """

#     # TODO: consider making this more flexible
#     # Example: forward(*x, loss: ModLoss)
#     @staticmethod
#     def forward(ctx, x: torch.Tensor, loss: DXModLoss):
#         with torch.enable_grad():
#             x = utils.freshen(x)
#             y = loss.module.forward(x)
#         ctx.save_for_backward(x, y)
#         ctx.loss = loss
#         return y

#     @staticmethod
#     def backward(ctx, dx: torch.Tensor):

#         with torch.enable_grad():
#             x, y = ctx.saved_tensors
#             result = ctx.loss.forward(x, y, dx)
#             result.backward()

#         return x.grad, None


# class TLossGrad(torch.autograd.Function):
#     """Use to implement backpropagation that is closer to the standard
#     form that sends information in a direct path backward through the layers
#     """

#     # TODO: consider making this more flexible
#     # Example: forward(*x, loss: ModLoss)
#     @staticmethod
#     def forward(ctx, x: torch.Tensor, loss: TModLoss):
#         with torch.enable_grad():
#             x = utils.freshen(x)
#             y = loss.module.forward(x)
#         ctx.save_for_backward(x, y)
#         ctx.loss = loss
#         return y

#     @staticmethod
#     def backward(ctx, dx: torch.Tensor):

#         with torch.enable_grad():
#             x, y = ctx.saved_tensors
#             result = ctx.loss.forward(x, y, y - dx)
#             result.backward()

#         return x.grad, None



"""

"""

import typing

# 1st Party
from abc import abstractmethod, abstractproperty
from dataclasses import dataclass
from enum import Enum

import numpy as np

# 3rd Party
import torch
import torch.nn as nn

# Local


class Reduction(Enum):

    MEAN = "mean"
    SUM = "sum"
    NONE = "none"
    BATCHMEAN = "batchmean"
    SAMPLEMEANS = "samplemeans"
    SAMPLESUMS = "samplesums"

    @classmethod
    def is_torch(cls, reduction: str):
        return reduction in ("mean", "sum", "none")

    def reduce(self, loss: torch.Tensor, dim=None):
        return Reduction.reduce_by(loss, self.value, dim)

    @classmethod
    def sample_reduce_by(
        cls,
        loss: torch.Tensor,
        reduction: typing.Union["Reduction", str],
        dim=None,
        keepdim: bool = False,
    ) -> torch.Tensor:

        view = torch.Size([loss.size(0), -1])
        if reduction == cls.MEAN.value:
            return loss.view(view).mean(1)
        if reduction == cls.SUM.value:
            return loss.view(view).sum(1)
        raise ValueError(f"{reduction} is an invalid reduction.")

    @classmethod
    def reduce_by(
        cls,
        loss: torch.Tensor,
        reduction: typing.Union["Reduction", str],
        dim=None,
        keepdim: bool = False,
    ) -> torch.Tensor:
        if isinstance(reduction, Reduction):
            reduction = reduction.value
        if reduction is None:
            reduction = "none"

        if reduction == cls.MEAN.value and dim is None:
            return loss.mean()
        elif reduction == cls.SUM.value and dim is None:
            return loss.sum()
        elif reduction == cls.MEAN.value:
            return loss.mean(dim=dim, keepdim=keepdim)
        elif reduction == cls.SUM.value:
            return loss.sum(dim=dim, keepdim=keepdim)
        elif reduction == cls.BATCHMEAN.value and dim is None:
            return loss.sum() / loss.size(0)
        elif reduction == cls.BATCHMEAN.value:
            return loss.sum(dim=dim, keepdim=keepdim) / loss.size(0)
        elif reduction == cls.SAMPLEMEANS.value:
            if loss.dim() == 1:
                return loss
            return loss.reshape(loss.size(0), -1).mean(dim=1, keepdim=keepdim)
        elif reduction == cls.SAMPLESUMS.value:
            if loss.dim() == 1:
                return loss
            return loss.reshape(loss.size(0), -1).sum(dim=1, keepdim=keepdim)
        elif reduction == cls.NONE.value:
            return loss
        raise ValueError(f"{reduction} is an invalid reduction.")


@dataclass
class Assessment(object):

    value: torch.Tensor
    maximize: bool = False

    def __getitem__(self, key):
        return Assessment(self.value[key], self.maximize)

    def __len__(self) -> int:
        return len(self.value) if self.value.dim() != 0 else 0

    def update_direction(self, maximize: bool):
        if self.maximize == maximize:
            return Assessment(self.value, maximize)
        return Assessment(-self.value, True)

    def __add__(self, other: "Assessment"):
        other = other.update_direction(self.maximize)
        return Assessment(self.value + other.value, self.maximize)

    def __sub__(self, other: "Assessment"):
        other = other.update_direction(self.maximize)
        return Assessment(self.value - other.value, self.maximize)

    def __mul__(self, val: float):
        return Assessment(self.value * val, self.maximize)

    def mean(self, dim=None) -> "Assessment":
        if dim is None:
            return Assessment(self.value.mean(), self.maximize)
        return Assessment(self.value.mean(dim=dim), self.maximize)

    def sum(self, dim=None) -> "Assessment":
        if dim is None:
            return Assessment(self.value.sum(), self.maximize)
        return Assessment(self.value.sum(dim=dim), self.maximize)

    def batch_mean(self) -> "Assessment":
        return Assessment(self.value.view(self.shape[0], -1).mean(dim=-1))

    def batch_sum(self) -> "Assessment":
        return Assessment(self.value.view(self.shape[0], -1).sum(dim=-1))

    def detach(self) -> "Assessment":
        return Assessment(self.value.detach(), self.maximize)

    def cpu(self) -> "Assessment":
        return Assessment(self.value.cpu(), self.maximize)

    def numpy(self) -> np.ndarray:
        return self.value.numpy()

    def item(self) -> typing.Any:
        if self.value.dim() != 0:
            raise ValueError(
                "Only one element tensors can be converted to Python scalars."
            )

        return self.value.item()

    def expand_population(self, population_size) -> "Assessment":
        return Assessment(
            self.value.view(
                population_size, self.shape[0] // population_size, *self.shape[1:]
            ),
            self.maximize,
        )

    def _extract_dim_greater_than_2(self, x: torch.Tensor, best_idx: torch.LongTensor):

        x_reshaped = x.view(x.size(0), x.size(1), -1)
        x_reshaped_index = best_idx[:, :, None].repeat(1, 1, x_reshaped.size(2))

        best_x = x_reshaped.gather(dim=0, index=x_reshaped_index)
        best_x = best_x.view(*x.shape[1:])

        # TODO: Possible to improve performance by not doing gather when creating batch assessment
        batch_assessment = Assessment(
            self.value.gather(dim=0, index=best_idx).flatten(), maximize=self.maximize
        )
        return best_x, batch_assessment, best_idx.flatten()

    def _extract_dim_is_2(self, x: torch.Tensor, best_idx: torch.LongTensor):
        assert best_idx.dim() == 1
        best_idx = best_idx[0]
        return (
            x[best_idx],
            Assessment(self.value[best_idx].flatten(), maximize=self.maximize),
            best_idx,
        )

    def extract_best(
        self, x: torch.Tensor
    ) -> typing.Tuple[torch.Tensor, "Assessment", torch.Tensor]:
        """Extract the best in each group value indices of the best in each batch of the population"""
        best_value, best_idx = (
            self.value.max(dim=0, keepdim=True)
            if self.maximize
            else self.value.min(dim=0, keepdim=True)
        )
        if x.dim() > 2:
            return self._extract_dim_greater_than_2(x, best_idx)

        return self._extract_dim_is_2(x, best_idx)

    def best(
        self, dim: int = 0, keepdim: bool = False
    ) -> typing.Tuple[torch.Tensor, torch.LongTensor]:

        if self.maximize:
            return self.value.max(dim=dim, keepdim=keepdim)
        return self.value.min(dim=dim, keepdim=keepdim)

    @property
    def shape(self) -> torch.Size:
        return self.value.shape

    def backward(self):
        if self.value.dim() != 0:
            raise ValueError("Backward can only be computed for  one element tensors .")

        return self.value.backward()

    def reduce(self, reduction: str = "mean", dim: int = None, keepdim: bool = False):
        """

        Args:
            reduction (str, optional): The reduction to apply. Defaults to 'mean'.
            dim (int, optional): The dim to apply the reduction on - only for mean and sum. Defaults to None.

        Returns:
            Assessment: The reduced assessment
        """
        return Assessment(
            Reduction.reduce_by(self.value, reduction, dim, keepdim=keepdim),
            self.maximize,
        )

    @classmethod
    def stack(self, assessments: typing.Iterable["Assessment"]):
        maximize = None
        values = []
        for assessment in assessments:
            if maximize is None:
                maximize = assessment.maximize
            if maximize != assessment.maximize:
                raise ValueError(
                    f"In order to stack optimizations must be in same direction. "
                    f"Current is {maximize} previous was {assessment.maximize}"
                )
            values.append(assessment.value)
        return Assessment(torch.stack(values), maximize)

    def to_dict(self, name: str) -> "AssessmentDict":
        return AssessmentDict(**{name: self})

    def __iter__(self) -> typing.Iterator["Assessment"]:

        if self.value.dim() == 0:
            yield Assessment(self.value, self.maximize)
        else:
            for element in self.value:
                yield Assessment(element, self.maximize)

    def view(self, *size: int):
        return Assessment(self.value.view(size), self.maximize)

    def reshape(self, *size: int):
        return Assessment(self.value.reshape(size), self.maximize)


def reduce_assessment(
    evaluation: torch.Tensor,
    maximize: bool = False,
    reduction: str = "mean",
    dim: int = -1,
):

    return Assessment(evaluation, maximize).reduce(reduction, dim)


class AssessmentDict(object):
    def __init__(self, **assessments: Assessment):
        self._assessments: typing.Dict[str, Assessment] = assessments

    def as_dict(self, to_tensor: bool = True):
        return {
            key: assessment.value if to_tensor else assessment
            for key, assessment in self._assessments.items()
        }

    def items(self) -> typing.Iterator[typing.Tuple[str, Assessment]]:

        for key, value in self._assessments.items():
            yield key, value

    def keys(self) -> typing.Iterator[str]:

        for key in self._assessments.keys():
            yield key

    def values(self) -> typing.Iterator[torch.Tensor]:
        """
        Returns:
            typing.Iterator[torch.Tensor]: Iterator for all of the values

        Yields:
            Iterator[typing.Iterator[torch.Tensor]]: A value
        """

        for value in self._assessments.values():
            yield value

    def __setitem__(self, key: str, value: Assessment):
        self._assessments[key] = value

    def __getitem__(
        self, key: typing.Union[typing.List[str], str]
    ) -> typing.Union[Assessment, "AssessmentDict"]:
        """
        Returns:
            Union[Assessment, 'AssessmentDict']: If input is a list, returns an AssessmentDict. Otherwise a string
        """

        if isinstance(key, typing.List):
            return AssessmentDict(
                **{
                    k: assessment
                    for k, assessment in self._assessments.items()
                    if k in key
                }
            )

        return self._assessments[key]

    def __len__(self) -> int:
        return len(self._assessments)

    def __contains__(self, key: str) -> bool:
        """
        Returns:
            bool: True if key is in the assessments
        """
        return key in self._assessments

    def _loop_pair(self, other: "AssessmentDict"):
        keys = set(self.keys()).union(set(other.keys()))
        for k in keys:
            yield k, self._assessments.get(k), other._assessments.get(k)

    def __add__(self, other: "AssessmentDict"):
        result = AssessmentDict()
        for key, s1, o1 in self._loop_pair(other):
            if s1 is not None and o1 is not None:
                result[key] = s1 + o1
            elif s1 is not None:
                result[key] = s1
            else:
                result[key] = o1

        return result

    def __sub__(self, other: "AssessmentDict") -> "AssessmentDict":
        """Take the difference in keys between the two assessment dicts

        Args:
            other (AssessmentDict)

        Returns:
            AssessmentDict: The resulting assessment dict
        """
        result = AssessmentDict()
        for key, s1, o1 in self._loop_pair(other):
            if s1 is not None and o1 is not None:
                result[key] = s1 - o1
            elif s1 is not None:
                result[key] = s1

        return result

    def __mul__(self, val: float):
        return AssessmentDict(**{k: v * val for k, v in self._assessments.items()})

    def union(
        self, assessment_dict: "AssessmentDict", override: bool = False
    ) -> "AssessmentDict":

        if not override:
            return AssessmentDict(**self._assessments, **assessment_dict._assessments)
        else:
            assessments = {**self._assessments}
            assessments.update(assessment_dict._assessments)
            return AssessmentDict(**assessments)

    def sum(self, *fields: str, dim: int = None) -> "AssessmentDict":
        result = {}
        fields = set(fields)

        for key, value in self._assessments.items():
            if key in fields:
                result[key] = value.sum(dim)
            elif key in fields:
                result[key] = Assessment(value.value.sum(dim=dim), value.maximize)
            else:
                result[key] = self._assessments[key]
        return AssessmentDict(**result)

    def mean(self, *fields: str, dim: int = None) -> "AssessmentDict":
        result = {}
        fields = set(fields)

        for key, value in self._assessments.items():
            if key in fields:
                result[key] = value.mean(dim)
            else:
                result[key] = self._assessments[key]
        return AssessmentDict(**result)

    def detach(self) -> "AssessmentDict":
        """
        Returns:
            AssessmentDict: assessment dict with all values detached
        """

        return AssessmentDict(**{key: value.detach() for key, value in self.items()})

    def cpu(self) -> "AssessmentDict":
        """
        Returns:
            AssessmentDict: assessment dict that has been converted to cpu
        """

        return AssessmentDict(**{key: value.cpu() for key, value in self.items()})

    def numpy(self) -> typing.Dict[str, np.ndarray]:
        """
        Returns:
            typing.Dict[str, np.ndarray]: dictionary with all values converted to numpy
        """
        return {key: value.numpy() for key, value in self.items()}

    def item(self) -> typing.Dict[str, typing.Any]:
        return {key: value.item() for key, value in self.items()}

    def backward(self, key: str):
        """
        Args:
            key (str): assessment to go backward on

        Raises:
            KeyError: if key is not in the assessments
        """
        if key not in self._assessments:
            raise KeyError(f"Key {key} is not in assessments.")
        return self._assessments[key].backward()

    def prepend(self, with_str: str) -> "AssessmentDict":
        """
        Args:
            with_str (str): string to prepend each value by

        Returns:
            AssessmentDict: AssessmentDict with all names prepended
        """
        return AssessmentDict(
            {with_str + key: value for key, value in self._assessments}
        )

    def append(self, with_str: str) -> "AssessmentDict":
        """
        Args:
            with_str (str): string to append each name with

        Returns:
            AssessmentDict:  AssessmentDict with all names appended
        """
        return AssessmentDict(
            {key + with_str: value for key, value in self._assessments}
        )

    @classmethod
    def stack(
        self, assessment_dicts: typing.Iterable["AssessmentDict"]
    ) -> "AssessmentDict":
        """Stack all of the assessments in the assessment dicts

        Args:
            assessment_dicts (typing.Iterable[AssessmentDict]): List of assessment dicts containing
            assessments to stack

        Returns:
            AssessmentDict: The assessment dict with stacked assessments
        """
        values = {}
        for assessment in assessment_dicts:

            for k, v in assessment.items():
                if k in values:
                    values[k].append(v)
                else:
                    values[k] = [v]
        return AssessmentDict(**{k: Assessment.stack(v) for k, v in values.items()})

    def transfer(
        self, source: str, destination: str, remove_source: bool = False
    ) -> "AssessmentDict":
        """Copy a reference to an assessment loss to another key in the AssessmentDict
        Args:
            source (str): name of loss to transfer
            destination (str): name of loss to transfer to

        Returns:
            AssessmentDict: self
        """

        self[destination] = self[source]
        if remove_source:
            del self._assessments[source]
        return self


class Loss(nn.Module):
    def __init__(self, reduction: str = "mean", maximize: bool = False):
        """
        Args:
            reduction (str, optional): Reduction to reduce by. Defaults to 'mean'.
            maximize (bool, optional): Whether to maximize or minimize. Defaults to False.
        """
        super().__init__()
        self.reduction = reduction
        self._maximize = maximize

    @property
    def maximize(self) -> bool:
        """
        Returns:
            bool: whether to maximize
        """
        return self._maximize

    def reduce(
        self, loss: torch.Tensor, reduction_override: str = None
    ) -> torch.Tensor:
        """Reduce the loss that is passed in
        Args:
            loss (torch.Tensor): the loss to reduce
            reduction_override (str, optional): whether to override the dfault reduction. Defaults to None.

        Returns:
            torch.Tensor: the reduced loss
        """
        return Reduction.reduce_by(loss, reduction_override or self.reduction)

    def assess(
        self, x: torch.Tensor, t: torch.Tensor, reduction_override: str = None
    ) -> Assessment:
        """Calculate the assessment of

        Args:
            x (torch.Tensor):
            t (torch.Tensor):
            reduction_override (str, optional): _description_. Defaults to None.

        Returns:
            Assessment: _description_
        """
        return Assessment(self.forward(x, t, reduction_override), self._maximize)

    def assess_dict(
        self,
        x: torch.Tensor,
        t: torch.Tensor,
        reduction_override: str = None,
        name: str = "loss",
    ):
        return AssessmentDict(
            **{
                name: Assessment(
                    self(x, t, reduction_override=reduction_override), self._maximize
                )
            }
        )

    @abstractmethod
    def forward(self, x: torch.Tensor, t: torch.Tensor, reduction_override: str = None):
        pass


LOSS_MAP = {
    "mse": nn.MSELoss,
    "bce": nn.BCELoss,
    "bce_with_logits": nn.BCEWithLogitsLoss,
    "cross_entropy": nn.CrossEntropyLoss,
}


class ThLoss(Loss):
    """ """

    def __init__(
        self,
        base_loss: typing.Union[typing.Callable[[str], nn.Module], str],
        reduction: str = "mean",
        weight: float = None,
        loss_kwargs: typing.Dict = None,
    ):
        """_summary_

        Args:
            base_loss (typing.Union[typing.Callable[[str], nn.Module], str]): _description_
            reduction (str, optional): _description_. Defaults to 'mean'.
            weight (float, optional): _description_. Defaults to None.
            loss_kwargs (typing.Dict, optional): _description_. Defaults to None.

        Raises:
            KeyError: _description_
        """
        """
        """

        # """Create a Loss that wraps a
        # Args:
        #     base_loss (typing.Type[nn.Module]): factory to create a torch loss
        #     base_reduction (str, optional): The default reduction to use. Defaults to 'mean'.
        # """
        super().__init__(reduction)
        if isinstance(base_loss, str):
            try:
                base_loss = LOSS_MAP[base_loss]
            except KeyError:
                raise KeyError(f"No loss named {base_loss} in loss keys")
        assert base_loss is not None
        self.base_loss = base_loss
        self._loss_kwargs = loss_kwargs or {}
        self._weight = weight

    def add_weight(self, evaluation: torch.Tensor):
        return evaluation * self._weight if self._weight is not None else evaluation

    def forward(self, x: torch.Tensor, t: torch.Tensor, reduction_override: str = None):
        reduction = reduction_override or self.reduction
        if Reduction.is_torch(reduction):
            # use built in reduction
            return self.add_weight(
                self.base_loss(reduction=reduction, **self._loss_kwargs).forward(x, t)
            )

        return self.add_weight(
            self.reduce(
                self.base_loss(reduction="none", **self._loss_kwargs).forward(x, t),
                reduction,
            )
        )


class ModLoss(Loss):
    """Wraps a module and a loss together so that more advanced backpropagation
    can be implemented
    """

    @abstractproperty
    def module(self) -> nn.Module:
        raise NotImplementedError

    @abstractmethod
    def forward(
        self, x: torch.Tensor, y: torch.Tensor, t: torch.Tensor, reduction_override=None
    ):
        raise NotImplementedError

    def assess(
        self,
        x: torch.Tensor,
        y: torch.Tensor,
        t: torch.Tensor,
        reduction_override: str = None,
    ):
        return Assessment(self(x, y, t, reduction_override), self._maximize)

    def assess_dict(
        self,
        x: torch.Tensor,
        y: torch.Tensor,
        t: torch.Tensor,
        reduction_override: str = None,
        name: str = "loss",
    ):
        return AssessmentDict(
            **{name: Assessment(self(x, y, t, reduction_override), self._maximize)}
        )


class ThModLoss(ModLoss):
    """Use for losses that implement the x, y, t interface for forward, assess, assess dict"""

    def __init__(
        self,
        base_loss_factory: typing.Callable[[str], ModLoss],
        reduction: str = "mean",
        weight: float = None,
        loss_kwargs: typing.Dict = None,
    ):
        """

        Args:
            base_loss (typing.Type[TModLoss]): s
            base_reduction (str, optional): . Defaults to 'mean'.
        """
        super().__init__(reduction)
        self.base_loss = base_loss_factory(reduction="none")
        self._loss_kwargs = loss_kwargs or {}
        self._weight = weight

    def forward(
        self, x: torch.Tensor, y: torch.Tensor, t: torch.Tensor, reduction_override=None
    ):
        return self.reduce(self.base_loss(x, y, t), reduction_override)



# 1st party
import typing
import uuid
from dataclasses import dataclass

# 3rd party
import torch
import torch.nn as nn

# TODO:
# Think more about this. I do not necessarily want to store
# the outputs (y values).
# If I store the outputs, for the nested version, I should
# store the intermediate evaluations as well.
#
# When the nested version is used, it will loop over the
# forward_iter if the inputs are defined else it will not
#


class ID(object):
    def __init__(self, id: uuid.UUID = None):

        self.x = id if id is not None else uuid.uuid4()


class _UNDEFINED:
    def __str__(self):
        return "UNDEFINED"


UNDEFINED = _UNDEFINED()


# TODO
class Func(nn.Module):
    """Module that wraps a function call"""

    def __init__(self, f: typing.Callable, *args, **kwargs):
        super().__init__()
        self._f = f
        self.args = args
        self.kwargs = kwargs

    def forward(self, x):
        return self._f(x, *self.args, **self.kwargs)


class Gen(nn.Module):
    """ """

    def __init__(self, generator: typing.Callable[[], torch.Tensor], *args, **kwargs):
        super().__init__()
        self._f = generator
        self.args = args
        self.kwargs = kwargs

    def forward(self, x: bool):

        if x is True:
            return self._f(*self.args, **self.kwargs)
        return UNDEFINED


@dataclass
class Info:
    tags: typing.List[str] = None
    annotation: str = None



# 1st party
from abc import ABC, abstractmethod

# local
from .assess import AssessmentDict


class NNComponent(ABC):
    """Base class for component. Use to build up a Learning Machine"""

    def is_(self, component_cls):
        if isinstance(self, component_cls):
            return True


class Learner(NNComponent):
    """Update the machine parameters"""

    @abstractmethod
    def learn(self, x, t) -> AssessmentDict:
        """Function for learning the mapping from x to t

        Args:
            x ([type]): The input values
            t ([type]): The target values to map to
        """
        raise NotImplementedError

    @abstractmethod
    def test(self, x, t) -> AssessmentDict:
        """Function for evaluating the mapping from x to t

        Args:
            x ([type]): The input values
            t ([type]): The target values to map to
        """
        raise NotImplementedError


class SelfLearner(NNComponent):
    """Update the machine parameters"""

    @abstractmethod
    def learn(self, x, y=None) -> AssessmentDict:
        """Function for learning the mapping from x to t

        Args:
            x ([type]): The input values
            t ([type]): The target values to map to
        """
        raise NotImplementedError

    @abstractmethod
    def test(self, x, y=None) -> AssessmentDict:
        """Function for evaluating the mapping from x to t

        Args:
            x ([type]): The input values
            t ([type]): The target values to map to
        """
        raise NotImplementedError


class Regressor(NNComponent):
    """Output a real value"""

    @abstractmethod
    def regress(self, x):
        raise NotImplementedError


class Classifier(NNComponent):
    """Output a categorical value"""

    @abstractmethod
    def classify(self, x):
        raise NotImplementedError


class Encoder(NNComponent):
    """Output a categorical value"""

    @abstractmethod
    def encode(self, x):
        raise NotImplementedError


class Decoder(NNComponent):
    """Output a categorical value"""

    @abstractmethod
    def decode(self, x):
        raise NotImplementedError


class Autoencoder(Encoder, Decoder):
    def reconstruct(self, x):
        return self.decode(self.encode(x))


class Assessor(NNComponent):
    @abstractmethod
    def assess(self, x, t, reduction_override: str = None) -> AssessmentDict:
        pass



# 1st part
import math
import typing
from abc import ABC, abstractmethod

# 3rd party
import torch
from torch.nn.parameter import Parameter

# local
from ..kaku import Assessor
from .core import (
    Individual,
    Population,
    binary_prob,
    cat_params,
    deflatten,
    expand,
    flatten,
)


class Populator(ABC):
    """Base class for creating a population from an individual"""

    @abstractmethod
    def __call__(self, individual: Individual) -> Population:
        """Spawn a population from an individual

        Args:
            individual (Individual): The individual to populate based on

        Returns:
            Population: The resulting population
        """
        pass

    @abstractmethod
    def spawn(self) -> "Populator":
        """Spawn a new populator from the current populator

        Returns:
            Populator: The spawned populator
        """
        pass


class StandardPopulator(Populator):
    """Populator that uses a standard populator method for all values in the individual"""

    @abstractmethod
    def populate(
        self, key: str, val: typing.Union[torch.Tensor, Parameter]
    ) -> typing.Union[torch.Tensor, Parameter]:
        pass

    def __call__(self, individual: Individual) -> Population:
        """Call the populate method for each value and spawn the population

        Args:
            individual (Individual): The individual to populate based no

        Returns:
            Population: The resulting population
        """
        expanded = {}
        for key, val in individual:
            cur = self.populate(key, val)
            if cur is not None:
                expanded[key] = cur
        return Population(**expanded)


class PopulatorDecorator(Populator):
    """Decorate the results of a populator"""

    def __init__(self, base_populator: Populator):
        """initializer

        Args:
            base_populator (Populator): The populator to decorate
        """
        self.base_populator = base_populator

    @abstractmethod
    def decorate(
        self, key: str, base_val, val: typing.Union[torch.Tensor, Parameter]
    ) -> typing.Union[torch.Tensor, Parameter]:
        """Decorate each value in the population

        Args:
            key (str): the key for the value in the dictionary
            base_val (): the value before populating
            val (typing.Union[torch.Tensor, Parameter]): The value after populating

        Returns:
            typing.Union[torch.Tensor, Parameter]: The result of the decoration
        """
        pass

    def __call__(self, individual: Individual) -> Population:
        """Spawn a population from an individual

        Args:
            individual (Individual): The individual to populate based on

        Returns:
            Population: The resulting population
        """
        populated = self.base_populator(individual)
        expanded = {}
        for key, val in populated:
            if key in individual:
                expanded[key] = self.decorate(key, individual[key], val)

        return Population(**expanded)

    @abstractmethod
    def spawn(self) -> "PopulatorDecorator":
        pass


class RepeatPopulator(StandardPopulator):
    """Populator that outputs all the same values for the population dimension"""

    def __init__(self, k: int):
        """initializer

        Args:
            k (int): The size of the population
        """
        self.k = k

    def populate(
        self, key: str, val: typing.Union[torch.Tensor, Parameter]
    ) -> typing.Union[torch.Tensor, Parameter]:
        """Expands each of the values by repeating along the population dimension

        Args:
            key (str): The name of the value
            val (typing.Union[torch.Tensor, Parameter]): the value to repeat

        Returns:
            typing.Union[torch.Tensor, Parameter]: The expanded value
        """
        return expand(val, self.k)

    def spawn(self) -> "RepeatPopulator":
        return RepeatPopulator(self.k)


def populate_t(t: torch.Tensor, k: int, name: str = "t") -> Population:
    """Convenience function to expand the t dimension along the population dimension

    Args:
        t (torch.Tensor): the tensor to expand
        k (int): the size of the population
        name (str, optional): the name of the value. Defaults to "t".

    Returns:
        Population: The result of the expansion
    """

    t = Individual(**{name: t})
    populator = RepeatPopulator(k)
    return populator(t)


class SimpleGaussianPopulator(StandardPopulator):
    """initializer"""

    def __init__(self, k: int):
        """initializer

        Args:
            k (int): The size of the population
        """
        self.k = k

    def populate(
        self, key: str, val: typing.Union[torch.Tensor, Parameter]
    ) -> typing.Union[torch.Tensor, Parameter]:
        """

        Args:
            key (str): The key for the element
            val (typing.Union[torch.Tensor, Parameter]): The element to create the population for

        Returns:
            typing.Union[torch.Tensor, Parameter]: The element expanded along the dimension
        """

        perturbation = torch.randn(self.k - 1, *val.shape, device=val.device)
        val = torch.cat([val[None], val[None] + perturbation])

    def spawn(self) -> "SimpleGaussianPopulator":
        return SimpleGaussianPopulator(self.k)


class GaussianPopulator(StandardPopulator):
    def __init__(self, k: int, std: float = 1, equal_change_dim: int = None):
        self.k = k
        self.std = std
        self.equal_change_dim = equal_change_dim

    def populate(self, key: str, val: torch.Tensor):

        shape = [self.k - 1, *val.shape]
        if self.equal_change_dim:
            shape[self.equal_change_dim] = 1
        noise = torch.randn(*shape, device=val.device) * self.std

        return torch.cat([val[None], val[None] + noise])

    def spawn(self) -> "GaussianPopulator":
        return GaussianPopulator(self.k, self.std, self.equal_change_dim)


class BinaryPopulator(StandardPopulator):
    def __init__(
        self,
        k: int = 1,
        keep_p: float = 0.1,
        equal_change_dim: int = None,
        to_change: typing.Union[int, float] = None,
        reorder_params: bool = True,
        zero_neg: bool = False,
    ):
        if 0.0 >= keep_p or 1.0 < keep_p:
            raise RuntimeError("Argument p must be in range (0.0, 1.0] not {keep_p}")
        assert k > 1
        self.keep_p = keep_p
        self.k = k
        self._equal_change_dim = equal_change_dim
        self._is_percent_change = isinstance(to_change, float)
        if self._is_percent_change:
            assert 0 < to_change <= 1.0
        elif to_change is not None:
            assert to_change > 0
        self._to_change = to_change
        self._reorder_params = reorder_params
        self._zero_neg = zero_neg

    # TODO: Move this to a "PopulationModifier" or a Decorator
    def _generate_keep(self, param: torch.Tensor):

        shape = [self.k - 1, *param.shape]
        if self._equal_change_dim is not None:
            shape[self._equal_change_dim] = 1

        param = (param > 0.0).type_as(param)
        keep = (torch.rand(*shape, device=param.device) < self.keep_p).type(param.dtype)

        if self._to_change is None:
            return keep

        if self._is_percent_change:
            ignore_change = (
                torch.rand(1, 1, *param.shape[1:], device=param.device)
                > self._to_change
            ).type_as(param)
        else:
            _, indices = torch.rand(
                math.prod(param.shape[1:]), device=param.device
            ).topk(self._to_change, dim=-1)
            ignore_change = torch.ones(math.prod(param.shape[1:]), device=param.device)
            ignore_change[indices] = 0.0
            ignore_change = ignore_change.view(1, 1, *param.shape[1:])

        return torch.max(keep, ignore_change)

    def populate(self, key: str, val: torch.Tensor):

        keep = self._generate_keep(val)

        changed = -val[None] if not self._zero_neg else (1 - val[None])
        perturbed_params = keep * val[None] + (1 - keep) * changed
        concatenated = cat_params(val, perturbed_params, reorder=True)
        if not self._reorder_params:
            return concatenated
        reordered = torch.randperm(len(concatenated), device=concatenated.device)
        return concatenated[reordered]

    def spawn(self) -> "BinaryPopulator":
        return BinaryPopulator(
            self.k,
            self.keep_p,
            self._equal_change_dim,
            self._to_change,
            self._reorder_params,
            self._zero_neg,
        )


class ConservativePopulator(PopulatorDecorator):
    def __init__(
        self,
        base_populator: Populator,
        percent_change: float = 0.1,
        same_change: bool = True,
    ):
        super().__init__(base_populator)
        if not (0.0 <= percent_change <= 1.0):
            raise ValueError("Percent change must be between 0 and 1")
        self.percent_change = percent_change
        self.same_change = same_change

    def decorate(
        self,
        key: str,
        base_val: torch.Tensor,
        val: typing.Union[torch.Tensor, Parameter],
    ) -> typing.Union[torch.Tensor, Parameter]:

        base_val = base_val[None]
        size = list(val.size())
        if self.same_change:
            size[0] = 1

        to_change = (
            torch.rand(*size, device=val.device) < self.percent_change
        ).type_as(val)
        return to_change * val + (1 - to_change) * base_val

    def spawn(self) -> "ConservativePopulator":
        return ConservativePopulator(
            self.base_populator.spawn(), self.percent_change, self.same_change
        )


class PerceptronProbPopulator(Populator):
    def __init__(
        self, learner: Assessor, k: int, x: str = "x", unactivated: str = "unactivated"
    ):
        self.k = k
        self.learner = learner
        self.unactivated = unactivated
        self.x = x

    def populate(
        self, key: str, val: typing.Union[torch.Tensor, Parameter]
    ) -> typing.Union[torch.Tensor, Parameter]:

        if key != self.unactivated:
            return None

        x_ = torch.clamp((val + 1) / 2, 0.001, 0.999)
        return (x_[None] < torch.rand(self.k, *x_.shape, device=x_.device)).type_as(
            x_
        ) * 2 - 1


class BinaryProbPopulator(Populator):
    def __init__(
        self,
        learner: Assessor,
        k: int,
        zero_neg: bool = True,
        loss_name: str = "loss",
        x: str = "x",
        t: str = "t",
    ):
        self.learner = learner
        self.k = k
        self.zero_neg = zero_neg
        self.loss_name = loss_name
        self.x = x
        self.t = t

    def generate_sample(
        self,
        base_size: torch.Size,
        dtype: torch.dtype,
        device=torch.device,
        prob: torch.Tensor = None,
    ):

        prob = prob or 0.5
        sample = torch.rand(self.k, *base_size, dtype=dtype, device=device)

        sample = (sample > prob).type_as(sample)
        if not self.zero_neg:
            sample = sample * 2 - 1
        return sample

    def __call__(self, x: Individual) -> Population:

        sample = self.generate_sample(x.size(), x.dtype, x.device)
        t = x[self.t]
        t = expand(t[0], self.k)
        sample = flatten(sample)
        assessment = self.learner.assess(sample, t, "none")["loss"]
        value = assessment.value[:, None]
        sample = sample.unsqueeze(sample.dim())
        value = deflatten(value, self.k)
        sample = deflatten(sample, self.k)
        prob = binary_prob(sample, value)
        sample = self.generate_sample(x.size(), x.dtype, x.device, prob)
        return Population(x=sample)

    def spawn(self) -> "BinaryProbPopulator":
        return BinaryProbPopulator(
            self.learner, self.k, self.zero_neg, self.loss_name, self.x, self.t
        )


class PopulationLimiter(object):
    def __call__(
        self,
        individual: Individual,
        population: Population,
        limit: torch.LongTensor = None,
    ) -> Population:

        if limit is None:
            return population

        result = {}

        for k, v in population:
            individual_v = individual[k][None]
            individual_v = individual_v.repeat(v.size(0), 1, 1)
            individual_v[:, :, limit] = v[:, :, limit].detach()
            result[k] = individual_v
        return Population(**result)



# flake8: noqa

from .core import ID, UNDEFINED, Func, Gen, Info
from .nodes import (
    Apply,
    End,
    In,
    Index,
    Joint,
    LambdaApply,
    Layer,
    Process,
    ProcessSet,
    ProcessVisitor,
    get_x,
    is_defined,
    to_incoming,
)
from .spawning import FSpawner, MSpawner, ProcessSpawner
from .tako import Filter, Nested, Network, Sequence, TagFilter, Tako, dive, layer_dive



# 1st party
import typing

# 3rd party
from torch.utils.data import Dataset
from tqdm import tqdm

# local
from ..kaku import Learner
from .base import Classroom, Desk, Teacher
from .materials import IODecorator, Material
from .reporting import Record, Results


class Trainer(Teacher):
    """A teacher that can train the learner"""

    def __init__(
        self,
        name: str,
        default_learner: typing.Union[Learner, str],
        default_material: typing.Union[Material, str],
        record: Record = None,
        desk: Desk = None,
        classroom: Classroom = None,
        window: int = 30,
    ):
        """initializer

        Args:
            name (str): The name of the teacher
            default_learner (typing.Union[Learner, str]): The default learning to train
            default_material (typing.Union[Material, str]): The default material to train with
            record (Record, optional): Record for the teacher to use. Defaults to None.
            desk (Desk, optional): Desk to use for retrieving materials and storing info. Defaults to None.
            classroom (Classroom, optional): The classroom to use. Defaults to None.
            window (int, optional): The size of the widnow for showing results. Defaults to 30.
        """

        super().__init__(name)
        self._desk = desk or Desk()
        self._classroom = classroom or Classroom()

        self.record = record or Record()
        self.learner = default_learner
        self.material = default_material
        self.window = window

    def teach(
        self,
        override_learner: typing.Union[Learner, str] = None,
        override_material: typing.Union[Dataset, str] = None,
        epoch: int = None,
        n_epochs: int = None,
    ):
        # learner = call_student(self._classroom, override_learner, self.learner)
        # material = get_material(self._desk, override_material, self.material)

        learner = self._classroom.get(override_learner) or self._classroom.get(
            self.learner
        )
        material = self._desk.get_material(
            override_material
        ) or self._desk.get_material(self.material)
        results = Results(self.window)
        logger = self.record.create_logger(self.name, material)
        with tqdm(total=len(material)) as pbar:
            for i, (x, t) in enumerate(material):
                assessment_dict = logger(learner.learn(x, t))
                self._assistants.assist(self._name, assessment_dict, (x, t))
                results.add(assessment_dict.numpy())
                aggregation = results.aggregate()
                if epoch is not None:
                    aggregation[
                        "Epoch"
                    ] = f'{epoch}/{"?" if n_epochs is None else n_epochs}'
                pbar.set_postfix(aggregation)
                pbar.update(1)


class Validator(Teacher):
    """A teacher that can test the learner"""

    def __init__(
        self,
        name: str,
        default_learner: typing.Union[Learner, str],
        default_material: typing.Union[Material, str],
        record: Record = None,
        desk: Desk = None,
        classroom: Classroom = None,
        show_progress: bool = True,
    ):
        """initializer

        Args:
            name (str): The name of the teacher
            default_learner (typing.Union[Learner, str]): The default learning to train
            default_material (typing.Union[Material, str]): The default material to train with
            record (Record, optional): Record for the teacher to use. Defaults to None.
            desk (Desk, optional): Desk to use for retrieving materials and storing info. Defaults to None.
            classroom (Classroom, optional): The classroom to use. Defaults to None.
            window (int, optional): The size of the widnow for showing results. Defaults to 30.
        """

        super().__init__(name)
        self._desk = desk or Desk()
        self._classroom = classroom or Classroom()

        self.record = record or Record()
        self.learner = default_learner
        self.material = default_material
        self.show_progress = show_progress

    def teach(
        self,
        override_learner: typing.Union[Learner, str] = None,
        override_material: typing.Union[Dataset, str] = None,
        epoch: int = None,
        n_epochs: int = None,
    ):

        learner = self._classroom.get(override_learner) or self._classroom.get(
            self.learner
        )
        material = self._desk.get_material(
            override_material
        ) or self._desk.get_material(self.material)

        # learner = call_student(self._classroom, override_learner, self.learner)
        # material = get_material(self._desk, override_material, self.material)
        results = Results(None)
        logger = self.record.create_logger(self._name, material)
        with tqdm(total=len(material)) as pbar:
            for i, (x, t) in enumerate(material):
                assessment_dict = logger(learner.test(x, t))
                self._assistants.assist(self._name, assessment_dict, (x, t))
                results.add(assessment_dict.numpy())
                aggregation = results.aggregate()
                if epoch is not None:
                    aggregation[
                        "Epoch"
                    ] = f'{epoch}/{"?" if n_epochs is None else n_epochs}'
                pbar.set_postfix(aggregation)
                pbar.update(1)


def validation_train(
    learner: Learner,
    training_material: Material,
    validation_material: Material,
    n_epochs: int = 1,
    use_io: bool = False,
) -> Record:

    record = Record()
    if use_io:
        training_material = IODecorator(training_material)
        validation_material = IODecorator(validation_material)
    trainer = Trainer("Trainer", learner, training_material, record=record)
    validator = Validator("Validator", learner, validation_material, record=record)

    for i in range(n_epochs):
        trainer.teach(epoch=i, n_epochs=n_epochs)
        validator.teach(epoch=i, n_epochs=n_epochs)
    return record


def train(
    learner: Learner,
    training_material: Material,
    testing_material: Material = None,
    n_epochs: int = 1,
    use_io: bool = False,
    window: int = 30,
) -> Record:

    record = Record()
    if use_io:
        training_material = IODecorator(training_material)
    trainer = Trainer("Trainer", learner, training_material, record=record, window=50)
    if testing_material:
        if use_io:
            testing_material = IODecorator(testing_material)
        tester = Validator("Validator", learner, testing_material, record=record)
    else:
        tester = None

    for i in range(n_epochs):
        trainer.teach(epoch=i, n_epochs=n_epochs)
    if tester is not None:
        tester.teach(epoch=i, n_epochs=n_epochs)
    return record



# 3rd party
import torch
import torch.nn as nn
import torch.nn.functional


class Lambda(nn.Module):
    """
    Use as a
    """

    def __init__(self, f):
        super().__init__()
        self._f = f

    def forward(self, *x: torch.Tensor):
        return self._f(*x)


class Argmax(nn.Module):
    def __init__(self, dim=-1):
        super().__init__()
        self._dim = dim

    def forward(self, x: torch.Tensor):
        return torch.argmax(x, dim=-1)


class Sign(nn.Module):
    def __init__(self, grad: bool = True):
        super().__init__()
        self._grad = grad

    def forward(self, x: torch.Tensor):
        if self._grad:
            return SignFG.apply(x)
        return torch.sign(x)


class SignFG(torch.autograd.Function):
    """Use to clip the grad between two values
    Useful for smooth maximum/smooth minimum
    """

    @staticmethod
    def forward(ctx, x: torch.Tensor):
        return torch.sign(x).type_as(x)

    @staticmethod
    def backward(ctx, grad: torch.Tensor):
        return grad


# ##### Least squares
# class LinearLayer(nn.Module):

#     def __init__(
#       self, in_features: int, out_features: int, bias: bool=True, 
#       in_activation: nn.Module=None, out_activation: nn.Module=None):

#         super().__init__()
#         self._in_features = in_features
#         self._out_features = out_features
#         self.linear = nn.Linear(in_features, out_features, bias=bias)
#         self.in_activation = in_activation
#         self.out_activation = out_activation

#     @property
#     def weight(self) -> nn.parameter.Parameter:
#         return self.linear.weight

#     @property
#     def bias(self) -> nn.parameter.Parameter:
#         return self.linear.bias

#     @property
#     def use_bias(self) -> bool:
#         return self.linear.bias is not None

#     def sub(self, idx: torch.LongTensor):
#         layer = LinearLayer(
#             self._in_features, len(idx), self.linear.bias is not None,
#             self.in_activation, self.out_activation
#         )
#         layer.linear.weight = nn.parameter.Parameter(self.linear.weight[idx])
#         layer.linear.bias = nn.parameter.Parameter(self.linear.bias[idx])
#         return layer

#     @property
#     def weight_with_bias(self) -> torch.Tensor:
#         if self.bias is not None:
#             return torch.cat(
#                 [self.weight, self.bias[:,None]], dim=1
#             )
#         return self.weight

#     def forward(self, x: torch.Tensor):
#         if self.in_activation is not None:
#             x = self.in_activation(x)
#         x = self.linear(x)
#         if self.out_activation is not None:
#             x = self.out_activation(x)
#         return x


# class BatchNorm1DS(nn.BatchNorm1d):

#     def forward(self, x: torch.Tensor, update_stats: bool=True):

#         if update_stats or not self.training:
#             normalized = super().forward(x)
#             print(normalized.mean(dim=0), normalized.std(dim=0))
#             return normalized

#         normalized = (x - x.mean(dim=0, keepdim=True)) / torch.sqrt(x.var(dim=0, keepdim=True) + self.eps)
#         if self.affine:
#             return normalized  * self.weight[None] + self.bias[None]
#         return normalized


# Rename
# class KLDivLossWithLogits(nn.Module):

#     def __init__(self, reduction: str='batchmean', class_target: bool=True):
#         super().__init__()
#         self._kl_div = nn.KLDivLoss(reduction=reduction)
#         self._class_target = class_target

#     def forward(self, x: torch.Tensor, t: torch.Tensor):
#         if self._class_target:
#             t = torch.nn.functional.one_hot(t)
#         return self._kl_div(nn_func.log_softmax(x, dim=1), t)


# # TODO: Remove if it does not work
# class BinaryStepAdv(torch.autograd.Function):
#     """Use to clip the grad between two values
#     Useful for smooth maximum/smooth minimum
#     """

#     @staticmethod
#     def forward(ctx, x):
#         """
#         Forward pass of the Binary Step function.
#         """
#         ctx.save_for_backward(x)
#         return torch.sign(x)

#     @staticmethod
#     def backward(ctx, grad_output):
#         """
#         Backward pass of the Binary Step function using the Straight-Through Estimator.
#         """
#         x, = ctx.saved_tensors
#         grad_input = grad_output.clone()
#         query = (
#             ((x < 1) & (grad_output > 0)) | ((x > 1) & (grad_output < 0))
#         )
#         grad_input[query] = 0
#         return grad_input



# 1st party
import typing

import numpy as np
import torch.nn as nn
import torch.nn.functional

# 3rd party
from sklearn.base import BaseEstimator
from sklearn.multioutput import MultiOutputClassifier, MultiOutputRegressor

# local
from .. import utils
from ..kaku import (
    IO,
    AssessmentDict,
    Conn,
    FeatureIdxStepTheta,
    FeatureIdxStepX,
    FeatureLimitGen,
    Idx,
    LearningMachine,
    Loss,
    State,
)
from ..utils import Argmax, Sign


class ScikitEstimator(nn.Module):
    """Wraps an Scikit Estimator in an nn.Module"""

    def __init__(
        self,
        sklearn_estimator,
        in_features: int = 1,
        out_features: int = 1,
        partial_fit: bool = True,
        use_predict: bool = True,
        backup: nn.Module = None,
    ):
        """initializer

        Args:
            sklearn_machine (_type_):
            multitarget (bool, optional): Whether or not the sklearn estimator is multitarget. Defaults to False.
            n_outputs (int, optional): The number of estimates . Defaults to 1.
            regressor (bool, optional): Whether the estimator is a regressor. Defaults to True.
            partial_fit (bool, optional): Whether to use partial fit. Defaults to True.
            use_predict (bool, optional): Whether to predict the output or use 'transform'. Defaults to True.
        """
        super().__init__()
        self._sklearn_estimator = sklearn_estimator
        if partial_fit and not hasattr(self._sklearn_estimator, "partial_fit"):
            raise RuntimeError(
                "Using partial fit but estimator does not have partial fit method available"
            )

        if not partial_fit and not hasattr(self._sklearn_estimator, "fit"):
            raise RuntimeError(
                "Using fit but estimator does not have fit method available"
            )

        self._output = self._predict if use_predict else self._transform
        self.fit = self._partial_fit if partial_fit else self._full_fit
        self._fitted = False
        self._in_features = in_features
        self._out_features = out_features
        self.backup = backup or nn.Linear(in_features, out_features)
        self._is_multioutput = isinstance(
            sklearn_estimator, MultiOutputClassifier
        ) or isinstance(sklearn_estimator, MultiOutputRegressor)

    def _predict(self, x):
        return self._sklearn_estimator.predict(x)

    def _transform(self, x):
        return self._sklearn_estimator.transform(x)

    def is_multioutput(self) -> bool:
        return self._is_multioutput

    @property
    def estimator(self) -> BaseEstimator:
        """
        Returns:
            BaseEstimator: the estimator wrapped by the SciKitEstimator
        """
        return self._sklearn_estimator

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Send the tensor through the estimator

        Args:
            x (torch.Tensor): the input

        Returns:
            torch.Tensor: the result of the scikit esimator converted to a Tensor
        """

        if not self.fitted:
            y = self.backup(x)
            return y
        x_np = utils.to_np(x)
        y = utils.to_th_as(self._sklearn_estimator.predict(x_np), x)

        return y

    def _prepare(
        self, y: np.ndarray, limit: typing.List[int]
    ) -> typing.List[BaseEstimator]:
        """

        Args:
            y (np.ndarray): _description_
            limit (typing.List[int]): _description_

        Returns:
            typing.List[BaseEstimator]: _description_
        """
        if limit is None:
            return y, None

        if not (
            isinstance(self._sklearn_estimator, MultiOutputClassifier)
            or isinstance(self._sklearn_estimator, MultiOutputRegressor)
        ):
            return y, None
            # raise ValueError(f"Cannot set limit if not using multioutput regressor or classifier")
        cur_estimators = self._sklearn_estimator.estimators_
        fit_estimators = [self._sklearn_estimator.estimators_[i] for i in limit]
        self._sklearn_estimator.estimators_ = fit_estimators
        return y[:, limit], cur_estimators

    def _replace_estimators(
        self, cur_estimators: typing.List[BaseEstimator], limit: typing.List[int]
    ):

        if limit is None or cur_estimators is None:
            return
        fit_estimators = self._sklearn_estimator.estimators_
        self._sklearn_estimator.estimators_ = cur_estimators
        for i, estimator in zip(limit, fit_estimators):
            self._sklearn_estimator.estimators_[i] = estimator

    def _partial_fit(
        self, X: torch.Tensor, y: torch.Tensor, limit: typing.List[int] = None
    ):
        if limit is not None and not self.fitted:
            raise RuntimeError("Must fit model before setting a limit")
        X = utils.to_np(X)
        y = utils.to_np(y)

        y, cur_estimators = self._prepare(y, limit)
        if limit is not None and len(limit) == 1:
            self._sklearn_estimator.partial_fit(X, y)
        else:
            self._sklearn_estimator.estimators_[0].partial_fit(X, y.flatten())

        self._replace_estimators(cur_estimators, limit)
        self.fitted = True

    def fit(self, X: torch.Tensor, y: torch.Tensor, limit: typing.List[int] = None):
        pass

    def _full_fit(
        self, X: torch.Tensor, y: torch.Tensor, limit: typing.List[int] = None
    ):
        """Runs a fit operation

        Args:
            X (torch.Tensor): the input
            y (torch.Tensor): the target tensor
            limit (typing.List[int], optional): the index of the limit. Defaults to None.

        Raises:
            RuntimeError: If the model has not been fit yet and a "limit" was set
        """
        if limit is not None and not self.fitted:
            raise RuntimeError("Must fit model before setting a limit")
        X = utils.to_np(X)
        y = utils.to_np(y)
        y, cur_estimators = self._prepare(y, limit)

        if limit is not None and len(limit) == 1:
            self._sklearn_estimator.estimators_[0].fit(X, y.flatten())
        else:
            self._sklearn_estimator.fit(X, y)
        self._replace_estimators(cur_estimators, limit)
        self._fitted = True

    @property
    def fitted(self) -> bool:
        """
        Returns:
            bool: If the model has been fitted already
        """
        return self._fitted


class ScikitRegressor(ScikitEstimator):
    def __init__(
        self,
        sklearn_estimator,
        in_features: int = 1,
        out_features: int = 1,
        multi: bool = False,
        partial_fit: bool = True,
        use_predict: bool = True,
    ):
        """_summary_

        Args:
            sklearn_machine (_type_):
            multitarget (bool, optional): Whether or not the sklearn estimator is multitarget. Defaults to False.
            n_outputs (int, optional): The number of estimates . Defaults to 1.
            regressor (bool, optional): Whether the estimator is a regressor. Defaults to True.
            partial_fit (bool, optional): Whether to use partial fit. Defaults to True.
            preprocessor (nn.Module, optional): . Defaults to None.
            postprocessor (nn.Module, optional): . Defaults to None.
            use_predict (bool, optional): Whether to predict the output or use 'transform'. Defaults to True.
        """
        if multi:
            sklearn_estimator = MultiOutputRegressor(sklearn_estimator)
        backup = nn.Linear(in_features, out_features)
        super().__init__(
            sklearn_estimator,
            in_features,
            out_features,
            partial_fit,
            use_predict,
            backup,
        )


class ScikitMulticlass(ScikitEstimator):
    """Adapter for a multiclass estimator"""

    def __init__(
        self,
        sklearn_estimator,
        in_features: int,
        n_classes: int,
        multi: bool = False,
        out_features: int = 1,
        partial_fit: bool = True,
        use_predict: bool = True,
        output_one_hot: bool = True,
    ):
        """_summary_

        Args:
            sklearn_machine : The estimator to adapt
            in_features (int): The number of features into the estimator
            n_classes (int): The number of classes to predict
            multi (bool, optional): Whether multioutput is used. Defaults to False.
            out_features (int, optional): The number of output features. Defaults to 1.
            partial_fit (bool, optional): Whether to use partial fit. Defaults to True.
            use_predict (bool, optional): Whether to use predict. Defaults to True.
            output_one_hot (bool, optional): Whether the output should be a one hot vector. Defaults to True.
        """

        if multi and out_features > 1:
            sklearn_estimator = MultiOutputClassifier(sklearn_estimator)

        backup = nn.Sequential(nn.Linear(in_features, n_classes), Argmax())
        self.output_one_hot = output_one_hot
        self._n_classes = n_classes
        super().__init__(
            sklearn_estimator,
            in_features,
            out_features,
            partial_fit,
            use_predict,
            backup,
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Args:
            x (torch.Tensor): the input to the estimator

        Returns:
            torch.Tensor: Output of the estimator convert to one hot if required
        """
        y = super().forward(x)

        if self.output_one_hot:
            return torch.nn.functional.one_hot(y, n_classes=self._n_classes)
        return y


class ScikitBinary(ScikitEstimator):
    """Adapter for a binary estimator"""

    def __init__(
        self,
        sklearn_estimator,
        in_features: int = 1,
        out_features: int = 1,
        multi: bool = False,
        partial_fit: bool = True,
        use_predict: bool = True,
    ):
        """initializer

        Args:
            sklearn_estimator (_type_): the estimator to adapt
            in_features (int, optional): The number of input features. Defaults to 1.
            out_features (int, optional): The number of output features. Defaults to 1.
            multi (bool, optional): Whether MultiOutput is used. Defaults to False.
            partial_fit (bool, optional): Whether to use partial_fit() or fit(). Defaults to True.
            use_predict (bool, optional): Whether to predict the output or use 'transform'. Defaults to True.
        """
        if multi and out_features > 1:
            sklearn_estimator = MultiOutputClassifier(sklearn_estimator)

        backup = nn.Sequential(nn.Linear(in_features, out_features), Sign())
        super().__init__(
            sklearn_estimator,
            in_features,
            out_features,
            partial_fit,
            use_predict,
            backup,
        )


class ScikitMachine(LearningMachine, FeatureIdxStepX, FeatureIdxStepTheta):
    """Machine used to train an estimator"""

    def __init__(
        self,
        module: ScikitEstimator,
        step_x: FeatureIdxStepX,
        loss: Loss,
        preprocessor: nn.Module = None,
    ):
        """initializer

        Args:
            module (ScikitEstimator): The
            step_x (FeatureIdxStepX): The function that does step_x
            loss (Loss): The loss function for the estimator
            preprocessor (nn.Module, optional): Module to preprocess the input sent to the estimator. Defaults to None.
        """
        super().__init__()
        self._module = module
        self._loss = loss
        self._step_x = step_x
        self._preprocessor = preprocessor

    def assess_y(self, y: IO, t: IO, reduction_override: str = None) -> AssessmentDict:
        return self._loss.assess_dict(y[0], t[0], reduction_override)

    def step(
        self, conn: Conn, state: State, from_: IO = None, feature_idx: Idx = None
    ) -> Conn:
        """Update the estimator

        Args:
            conn (Conn): Connection for the step
            state (State): The current state
            from_ (IO, optional): the input to the previous layer. Defaults to None.
            feature_idx (Idx, optional): . Defaults to None.

        Returns:
            Conn: the connection incremeented
        """
        x = conn.step.x[0]
        if self._preprocessor is not None:
            x = self._preprocessor(x)

        t = conn.step.t[0]
        self._module.fit(
            x, t, feature_idx.tolist() if feature_idx is not None else None
        )

        return conn.connect_in(from_)

    def step_x(self, conn: Conn, state: State, feature_idx: Idx = None) -> Conn:
        """Update the estimator

        Args:
            conn (Conn):
            state (State): The state of the
            feature_idx (Idx, optional): _description_. Defaults to None.

        Returns:
            Conn: the connection with an updated step target
        """
        return self._step_x.step_x(conn, state, feature_idx)

    def forward(self, x: IO, state: State, detach: bool = True) -> IO:
        """

        Args:
            x (IO): input to the machine
            state (State): the state
            detach (bool, optional): Whether to detach the output. Defaults to True.

        Returns:
            IO: output of the machine
        """

        x = x[0]
        if self._preprocessor is not None:
            x = self._preprocessor(x)

        y = IO(self._module(x))
        return y.out(detach=detach)

    @property
    def fitted(self) -> bool:
        return self._module.fitted


class ScikitLimitGen(FeatureLimitGen):
    """Use to generate a limit on the features that are trained
    Only use if the machine has been fit
    """

    def __init__(self, base_limiter: FeatureLimitGen):
        """initializer

        Args:
            base_limiter (FeatureLimitGen): The base limiter to use
        """
        self.base_limiter = base_limiter

    def sample_limit(self) -> Idx:
        return self.base_limiter()

    def __call__(self, fitted: bool = True) -> Idx:
        """Only executes sample_limit if the machine has been fitted

        Args:
            fitted (bool, optional): Whether the machine was fitted. Defaults to True.

        Returns:
            Idx: The limit on the output
        """
        if not fitted:
            return None

        return self.sample_limit()



# flake8: noqa

"""
Basic learning machine classes. 

A learning machine consists of parameters, operations and the interface
Mixins are used to flexibly define the interface for a learning machine
Each Mixin is a "machine component" which defines an interface for the
user to use. Mixins make it easy to reuse components.

class BinaryClassifierLearner(Learner, Tester, Classifier):
  
  def __init__(self, ...):
      # init operations

  def classify(self, x):
      # classify

  def learn(self, x, t):
      # update parameters of network
    
  def test(self, x, t):
      # evaluate the ntwork
    
  def forward(self, x):
      # standard forward method

"""

from ..kikai.hill import HillClimbBinaryStepX, HillClimbStepX
from .grad import (
    GradLearner,
    GradLoopLearner,
    GradLoopStepTheta,
    GradLoopStepX,
    GradStepTheta,
    GradStepX,
    NullStepTheta,
    update_x,
)
from .least_squares import (
    LeastSquaresLearner,
    LeastSquaresRidgeSolver,
    LeastSquaresSolver,
    LeastSquaresStandardSolver,
    LeastSquaresStepTheta,
    LeastSquaresStepX,
)
from .reversible import ReversibleMachine



"""
Factories for creating optimizers
"""

# 1st Party
import typing

# 3rd Party
import torch
import torch.nn as nn
import torch.nn.functional
import torch.optim as optim


class NullOptim(torch.optim.Optimizer):
    def __init__(self, parameters):
        self.state = {}

    def step(self):
        pass

    def state_dict(self) -> dict:
        return {}

    def load_state_dict(self, state_dict: dict) -> None:
        pass

    def zero_grad(self) -> None:
        pass


OPTIM_MAP = {
    "sgd": torch.optim.SGD,
    "adam": torch.optim.Adam,
    "adadelta": torch.optim.Adadelta,
    "adagrad": torch.optim.Adagrad,
    "rmsprop": torch.optim.RMSprop,
    "null": NullOptim,
}


OptimFactoryX = typing.Callable[[typing.Iterable], optim.Optimizer]


class OptimFactory(object):
    def __init__(
        self,
        optim: typing.Union[str, typing.Type[torch.optim.Optimizer]],
        *args,
        **kwargs,
    ):

        try:
            optim = OPTIM_MAP[optim] if isinstance(optim, str) else optim
        except KeyError:
            raise KeyError(f"No optim named {optim} in the optim map")
        self._optim = optim
        self._args = args
        self._kwargs = kwargs

    def __call__(self, params, **kwarg_overrides):

        kwargs = {**self._kwargs, **kwarg_overrides}
        return self._optim(params, *self._args, **kwargs)

    @classmethod
    def sgd(cls, *args, **kwargs):
        return OptimFactory(torch.optim.SGD, *args, **kwargs)

    @classmethod
    def adam(cls, *args, **kwargs):
        return OptimFactory(torch.optim.Adam, *args, **kwargs)

    @classmethod
    def adadelta(cls, *args, **kwargs):
        return OptimFactory(torch.optim.Adadelta, *args, **kwargs)

    @classmethod
    def asgd(cls, *args, **kwargs):
        return OptimFactory(torch.optim.ASGD, *args, **kwargs)

    @classmethod
    def rmsprop(cls, *args, **kwargs):
        return OptimFactory(torch.optim.RMSprop, *args, **kwargs)


class MetaOptim(optim.Optimizer):
    def __init__(
        self,
        p: typing.Iterable,
        meta_optim: OptimFactory,
        active_optim: OptimFactory = None,
        copy_first: bool = False,
    ):
        cloned = []
        base = []
        for p_i in p:
            cloned_i = torch.clone(p_i).detach().requires_grad_()
            cloned_i.retain_grad()
            cloned.append(cloned_i)
            base.append(p_i)

        self.meta_params = cloned
        self.active_params = base
        self.meta_optim = meta_optim(self.meta_params)
        active_optim = active_optim or NullOptim
        self.active_optim_factory = active_optim
        self.meta_optim_factory = meta_optim
        self.active_optim = active_optim(self.active_params)
        self.copy_first = copy_first
        self._is_first = True

    def step(self):
        self.active_optim.step()

    def zero_grad(self):
        self.active_optim.zero_grad()

    @property
    def state(self):
        return self.active_optim.state

    def state_dict(self):
        return {"active_params": self.active_params, "meta_params": self.meta_params}

    def load_state_dict(self, state_dict):

        self.meta_params = state_dict["meta_params"]
        self.active_params = state_dict["active_params"]
        self.meta_optim = self.meta_optim_factory(self.meta_params)
        self.active_optim = self.active_optim_factory(self.active_params)

    @property
    def param_groups(self):
        return self.active_optim.param_groups

    def add_param_group(self, param_group: dict):
        self.active_optim.add_param_group(param_group)
        self.meta_optim.add_param_group(param_group)

    def copy_meta_to(self, p: typing.Iterable):
        for p_i, mp_i in zip(p, self.meta_params):
            if isinstance(p_i, nn.parameter.Parameter):
                p_i.data = mp_i.data
            else:
                p_i[:] = mp_i.data

    def step_meta(self):
        self.meta_optim.zero_grad()

        if self._is_first and self.copy_first:
            for p_i, mp_i in zip(self.active_params, self.meta_params):
                if isinstance(p_i, nn.parameter.Parameter):
                    mp_i.data = p_i.data
                else:
                    mp_i.data[:] = p_i
        else:
            for active, meta in zip(self.active_params, self.meta_params):
                loss = (0.5 * (meta - active.detach()) ** 2).sum()
                loss.backward()
            self.meta_optim.step()

        self.meta_optim.zero_grad()
        self._is_first = False

    def transfer(self, clear_active_state: bool = True):
        for active, meta in zip(self.active_params, self.meta_params):
            active.data = meta.data

        if clear_active_state:
            self.active_optim.state.clear()
        # self.active_optim = self.active_optim_factory(self.active_params)

    def adv(self, clear_active_state: bool = True):
        self.step_meta()
        self.transfer(clear_active_state)


class itadaki:
    def __init__(self) -> None:
        raise RuntimeError("Cannot create instance of class itadaki")

    @staticmethod
    def adam(*args, **kwargs):
        return OptimFactory("adam", *args, **kwargs)

    @staticmethod
    def sgd(*args, **kwargs):
        return OptimFactory("sgd", *args, **kwargs)

    @staticmethod
    def null():
        return OptimFactory("null")



# 1st party
import typing

from ..kaku import IO, AssessmentDict, Conn, LearningMachine, State, ThLoss

# local
from ..utils import Reversible, SequenceReversible


class ReversibleMachine(LearningMachine):
    def __init__(
        self,
        reversible: typing.Union[Reversible, typing.List[Reversible]],
        loss: ThLoss,
        maximize: bool = False,
    ):
        """initializer

        Args:
            reversible (typing.Union[Reversible, typing.List[Reversible]]): Reversible module to adapt
            loss (ThLoss): The loss
            maximize (bool, optional): _description_. Defaults to False.
        """
        super().__init__(maximize)
        if isinstance(reversible, typing.List):
            reversible = SequenceReversible(*reversible)
        self.reversible = reversible
        self.loss = loss

    def assess_y(self, y: IO, t: IO, reduction_override: str = None) -> AssessmentDict:
        return self.loss.assess_dict(y, t, reduction_override).transfer(
            "loss", self.validation_name
        )

    def step_x(self, conn: Conn, state: State) -> Conn:
        """Update x

        Args:
            conn (Conn): The connection to update based on
            state (State): The learning state

        Returns:
            Conn: The connection with an updated target for step
        """
        conn.step_x.x_(self.reversible.reverse(conn.step_x.t[0]))
        conn.tie_inp(True)
        return conn

    def step(self, conn: Conn, state: State, from_: IO = None) -> Conn:
        """These layers do not have parameters so the internal mechanics are not updated

        Args:
            conn (Conn): The connection for the layer
            state (State): The learning state
            from_ (IO, optional): The input to the previous layer. Defaults to None.

        Returns:
            Conn: the connection for the preceding layer
        """
        return conn.connect_in(from_)

    def forward(self, x: IO, state: State, detach: bool = True) -> IO:
        return IO(self.reversible(x[0]), detach=detach)



# flake8: noqa

from .base import Assistant, Classroom, Desk, Material, Teacher
from .materials import DLMaterial, IODecorator, MaterialDecorator, split_dataset
from .reporting import Entry, Log, Logger, Record, Results
from .teaching import Trainer, Validator, train, validation_train



# local
from ..kaku.machine import (
    IO,
    Conn,
    FeatureIdxStepX,
    Idx,
    LearningMachine,
    State,
    update_io,
)
from ..tansaku.assessors import XPopulationAssessor
from ..tansaku.core import Individual
from ..tansaku.modifiers import SlopeModifier
from ..tansaku.populators import BinaryPopulator, GaussianPopulator, PopulationLimiter
from ..tansaku.selectors import BestSelectorFeature


class HillClimbStepX(FeatureIdxStepX):
    def __init__(
        self,
        learner: LearningMachine,
        k: int,
        std: float = 0.1,
        lr: float = 1e-2,
        momentum: float = 0.5,
        maximize: bool = False,
    ):
        """use a hill climbing algorithm to update the input values

        Args:
            learner (LearningMachine):
        """

        super().__init__()
        self.learner = learner
        self.limiter = PopulationLimiter()
        self.populator = GaussianPopulator(k, std=std)
        self.modifier = SlopeModifier(momentum, lr, maximize=maximize)
        self.assessor = XPopulationAssessor(self.learner, ["x"], "loss", "mean", k)

    def step_x(self, conn: Conn, state: State, feature_idx: Idx = None) -> Conn:

        individual = Individual(x=conn.step_x.x[0])

        population = self.limiter(
            individual,
            self.populator(individual),
            feature_idx.tolist() if feature_idx is not None else None,
        )
        population = self.assessor(population, conn.step_x.t)
        selected = self.modifier(individual, population)
        # conn.step_x.x_(IO(selected['x'], detach=True))
        update_io(IO(selected["x"], detach=True), conn.step_x.x)
        conn.tie_inp()
        return conn


class HillClimbBinaryStepX(FeatureIdxStepX):
    def __init__(self, learner: LearningMachine, k: int = 8, keep_p: float = 0.9):
        """use a hill climbing algorithm to update the input values

        Args:
            learner (LearningMachine):
        """
        super().__init__()
        self.learner = learner
        self.populator = BinaryPopulator(k, keep_p)
        self.selector = BestSelectorFeature()  # to_sample=False)
        self.limiter = PopulationLimiter()
        # self.modifier = tansaku.BinaryGaussianModifier(k_best, "x")
        self.assessor = XPopulationAssessor(self.learner, ["x"], "loss", "mean", k)

    @property
    def update_populator(self, k: int, keep_p: float):
        self.populator = BinaryPopulator(k, keep_p)

    def step_x(self, conn: Conn, state: State, feature_idx: Idx = None) -> Conn:

        individual = Individual(x=conn.step_x.x[0])
        # pass conn.limit into the populator
        # perhaps call this a kind of modifier?
        population = self.limiter(
            individual,
            self.populator(individual),
            feature_idx.tolist() if feature_idx is not None else None,
        )
        population = self.assessor(population, conn.step_x.t)
        selected = self.selector(population)
        # selected = self.modifier(individual, population)
        # conn.step_x.x_(IO(selected['x'], detach=True))
        update_io(IO(selected["x"], detach=True), conn.step_x.x)
        conn.tie_inp()
        return conn



import typing
from abc import ABC, abstractmethod

import torch
from torch.utils import data as torch_data

from .machine import (
    IO,
    BatchIdxStepTheta,
    BatchIdxStepX,
    Conn,
    Idx,
    LearningMachine,
    idx_conn,
    update_step_x,
)
from .state import State


class Step(ABC):
    """
    Base class for Steps. Steps are used to
    """

    @abstractmethod
    def step(self, conn: Conn, state: State, from_: IO = None) -> Conn:
        pass


class StepLoop(object):
    def __init__(self, batch_size: int, shuffle: bool = True):
        """Loop over a connection by indexing

        Args:
            batch_size (int): The size of the batch for the loop
            shuffle (bool, optional): whether to shuffle the indices. Defaults to True.
        """
        self.batch_size = batch_size
        self.shuffle = shuffle

    def create_dataloader(self, conn: Conn) -> torch_data.DataLoader:
        """
        Args:
            conn (Conn): the connection to create the dataloader for

        Returns:
            DataLoader: The data loader to loop over
        """

        batch_size = (
            self.batch_size if self.batch_size is not None else len(conn.inp.x[0])
        )

        # TODO: Change so 0 is not indexed
        indices = torch_data.TensorDataset(torch.arange(0, len(conn.out.x[0])))
        return torch_data.DataLoader(indices, batch_size, self.shuffle)

    def loop(self, conn: Conn) -> typing.Iterator[Conn]:
        """Loop over the connection

        Args:
            conn (Conn): the connection to loop over

        Returns:
            typing.Iterator[Conn]: _description_

        Yields:
            Iterator[typing.Iterator[Conn]]: _description_
        """
        for (idx,) in self.create_dataloader(conn):
            yield Idx(idx, dim=0)


class IterOutStep(Step):
    def __init__(
        self, learner: LearningMachine, n_epochs: int = 1, batch_size: int = None
    ):
        """
        Args:
            learner (LearningMachine): The LearningMachine to optimize
            n_epochs (int, optional): The number of epochs. Defaults to 1.
            batch_size (int, optional): . Defaults to None.
        """

        self.learner = learner
        self.n_epochs = n_epochs
        self.batch_size = batch_size

    def step(self, conn: Conn, state: State, from_: IO = None) -> Conn:
        """
        Args:
            conn (Conn):
            state (State):
            from_ (IO, optional): . Defaults to None.

        Returns:
            Conn: _description_
        """
        loop = StepLoop(self.batch_size, True)
        for _ in range(self.n_epochs):
            for idx in loop.loop(conn):

                # TODO: Consider how to handle this so I .
                # Use BatchIdxStep after all <- override the step method
                if isinstance(self.learner, BatchIdxStepTheta):
                    self.learner.step(conn, state, batch_idx=idx)
                else:
                    conn = idx_conn(conn)
                    self.learner.step(conn, state)

        return conn.connect_in(from_)


class IterHiddenStep(Step):
    def __init__(
        self,
        incoming: LearningMachine,
        outgoing: LearningMachine,
        n_epochs: int = 1,
        x_iterations: int = 1,
        theta_iterations: int = 1,
        x_batch_size: int = None,
        batch_size: int = None,
        tie_in_t: bool = True,
    ):
        """

        Args:
            incoming (LearningMachine): incoming layer
            outgoing (LearningMachine): outgoing layer
            n_epochs (int, optional): number of epochs. Defaults to 1.
            x_iterations (int, optional): . Defaults to 1.
            theta_iterations (int, optional): . Defaults to 1.
            x_batch_size (int, optional): . Defaults to None.
            batch_size (int, optional): . Defaults to None.
            tie_in_t (bool, optional): . Defaults to True.
        """
        self.incoming = incoming
        self.outgoing = outgoing
        self.n_epochs = n_epochs
        self.x_iterations = x_iterations
        self.theta_iterations = theta_iterations
        self.x_batch_size = x_batch_size
        self.batch_size = batch_size
        self.tie_in_t = tie_in_t

    def step(
        self,
        conn: Conn,
        state: State,
        from_: IO = None,
        clear_outgoing_state: bool = True,
    ):
        theta_loop = StepLoop(self.batch_size, True)
        x_loop = StepLoop(self.x_batch_size, True)

        for i in range(self.n_epochs):

            for _ in range(self.x_iterations):
                for idx in x_loop.loop(conn):
                    if isinstance(self.outgoing, BatchIdxStepX):
                        self.outgoing.step_x(conn, state, batch_idx=idx)
                    else:
                        conn_idx = idx_conn(conn, idx)
                        self.outgoing.step_x(conn_idx, state)
                        update_step_x(conn_idx, conn, idx, True)

            for _ in range(self.theta_iterations):

                for i, idx in enumerate(theta_loop.loop(conn)):
                    if isinstance(self.incoming, BatchIdxStepTheta):
                        self.incoming.step(conn, state, batch_idx=idx)
                    else:
                        # TODO: Add state into idx_conn (?)
                        conn_idx = idx_conn(conn, idx)
                        self.incoming.step(conn_idx, state)

            # TODO: Decide whether this is the default
            if self.tie_in_t and i < (self.n_epochs - 1):
                conn.inp.y_(self.incoming(conn.inp.x))
                conn.tie_out(True)

        # if in_step is None:
        #     raise ValueError(f'Could not loop over output with {self.incoming}')

        if clear_outgoing_state:
            state.clear(self.outgoing)
        return conn.connect_in(from_)


class TwoLayerStep(Step):
    def __init__(
        self,
        layer1: LearningMachine,
        layer2: LearningMachine,
        layer1_batch_size: int,
        layer2_batch_size: int,
        n_iterations: int = 1,
    ):
        super().__init__()
        self.layer1 = layer1
        self.layer2 = layer2
        self.layer1_batch_size = layer1_batch_size
        self.layer2_batch_size = layer2_batch_size
        self.n_iterations = n_iterations
        self.loop1 = StepLoop(self.layer1_batch_size, True)
        self.loop2 = StepLoop(self.layer2_batch_size, True)

    def step(self, x_in: IO, conn: Conn, state: State, from_: IO = None):

        for i in range(self.n_iterations):
            for conn_i in self.loop2.loop(conn):
                self.layer2.step(conn_i, state, from_=x_in)
            conn2 = conn.connect_in(x_in)
            conn2 = self.layer2.step_x(conn2, state)
            for conn2_i in self.loop1.loop(conn2):
                self.layer1.step(conn2_i, state, from_=from_)
            conn.step.x = self.layer1(conn2.step.x).detach()
        return conn2


# class Loop(ABC):

#     @abstractmethod
#     def loop(self, x, t) -> typing.Iterator:
#         pass


# class NullLoop(Loop):

#     def loop(self, x, t) -> typing.Iterator:
#         yield x, t


# class DLLoop(Loop):

#     def __init__(self, batch_size: int, shuffle: bool=True, get_indices: bool=False):
#         self.batch_size = batch_size
#         self.shuffle = shuffle
#         self.get_indices = get_indices

#     def create_dataloader(self, x: torch.Tensor, t: torch.Tensor):

#         if self.get_indices:
#             indices = torch.range(0, len(x))
#             dataset = torch_data.TensorDataset(x, t, indices)
#         else:
#             dataset = torch_data.TensorDataset(x, t)
#         return torch_data.DataLoader(
#             dataset, self.batch_size, self.shuffle
#         )

#     def loop(self, x, t) -> typing.Iterator:
#         for result in self.create_dataloader(x, t):
#             yield result
